# !/usr/bin/python # pylint: disable=I0011, C0302
# -*- coding: utf-8 -*-
r"""Core classes of natu

As shown in the diagram below, there are two types of units (:class:`Unit`):
scalar units (:class:`ScalarUnit`) and lambda units (:class:`LambdaUnit`).
Scalar units are quantities, but lambda units are not.  Lambda units are
essentially invertible functions that map between numbers and quantities
(:class:`Quantity`).  All units and quantities are dimensioned objects
(:class:`DimObject`), meaning that their physical dimension is recorded (even if
dimensionless).

.. inheritance-diagram:: DimObject Unit ScalarUnit LambdaUnit Quantity

Numbers, scalar units, and quantities can be used together in mathematical
expressions as long as the expression is dimensionally consistent.  Numbers
(:class:`float`, :class:`int`, :class:`~fractions.Fraction`, etc.) are
considered dimensionless.

Lambda units can only be used with numbers and quantities in certain ways.  A
lambda unit can be applied to generate a quantity for use in the rest of a
mathematical expression.  The multiplication operator (``*``) is overloaded so
that a number times a lambda unit calls the unit's method to map the number to a
quantity.  Likewise, the division operator (``/``) uses a lambda unit to map a
quantity to a number.

The tables below summarize the type of result given by simple mathematical
operations among numbers (N), scalar units (S), lambda units (L), and quantities
(Q).  The product or quotient of a number and a quantity or scalar unit is a
quantity.  This is the case even if the number is zero.  For example, the
product of zero and a unit still carries the dimension of the unit.  In a
Boolean expression, a quantity is cast as *False* if its value is zero.
However, it is only equal to zero if its value is zero and it is dimensionless.
The display unit of the first term in addition or subtraction takes precedence.
If the first term is a number and the second term is a dimensionless quantity,
then the result is a number.

Results of multiplication and division:

+--------------+---------------------------+
|              | 2nd argument              |
|              +----+--------+--------+----+
| 1st argument | N  | Q      | S      | L  |
+==============+====+========+========+====+
| **N**        | N  | Q      | Q      | Q  |
+--------------+----+--------+--------+----+
| **Q**        | Q  | Q [1]_ | Q [1]_ | N  |
+--------------+----+--------+--------+----+
| **S**        | Q  | Q [1]_ | S [1]_ | -- |
+--------------+----+--------+--------+----+
| **L**        | -- | --     | --     | -- |
+--------------+----+--------+--------+----+

Results of addition and subtraction:

+--------------+-------------------------------+
|              | 2nd argument                  |
|              +--------+--------+--------+----+
| 1st argument | N      | Q      | S      | L  |
+==============+========+========+========+====+
| **N**        | N      | N [2]_ | N [2]_ | -- |
+--------------+--------+--------+--------+----+
| **Q**        | Q [2]_ | Q      | Q      | -- |
+--------------+--------+--------+--------+----+
| **S**        | Q [2]_ | Q      | Q      | -- |
+--------------+--------+--------+--------+----+
| **L**        | --     | --     | --     | -- |
+--------------+--------+--------+--------+----+

Key:

- N: :class:`~numbers.Number`
- Q: :class:`Quantity`
- S: :class:`ScalarUnit`
- L: :class:`LambdaUnit`
- --: Not allowed

.. [1] If the result is dimensionless, then it is returned as a number instead.
.. [2] The quantity must be dimensionless.

Here are some consequences of this design:

- Is ``0*ppm`` equal to zero? Yes
- Is ``0*m`` equal to zero? No
- Does ``0*m`` evaluate to ``False``? Yes
- Is ``0*m`` dimensionless? No
- Is ``m/ft`` a quantity or a number? Number (float)
- Is ``ohm*S`` a quantity or a number? Number (1.0)
- Is ``m*m`` a unit? Yes (with display unit ``'m2'``)
- Is ``1*m*m`` a unit? No (only a quantity)
- Is ``1*ppm`` a unit or a number? Neither (a dimensionless quantity)
- Is ``m + 0`` (or ``0 + m``) valid? No
- Is ``ppm + 0`` (and ``0 + ppm``) valid? Yes
- Is ``ppm + 0`` a quantity or a number? Quantity (1 ppm)
- Is ``0 + ppm`` a quantity or a number? Number (1e-6)
- What is the display unit of ``ft + m``? ``'ft'``
- Is ``m*10`` valid? Yes (``Quantity(10, 'L', 'm') (10.0 m)``)
- Is ``degC*10`` valid? No (use ``10*degC`` instead)

Classes in this module:

- :class:`CoherentRelations` - List of coherent relations among units

- :class:`CompoundDimension` - Dictionary that contains base physical dimensions
  as keys and exponents of those dimensions as values

- :class:`CompoundUnit` - Dictionary that contains the base factors of a display
  unit as keys and exponents of those units as values

- :class:`DimObject` - Base class that records physical dimension and display
  unit

- :class:`LambdaUnit` - Unit that involves an offset or other operations besides
  scaling

- :class:`Quantity` - Class to represent a physical quantity

- :class:`ScalarUnit` - Unit that involves only a scaling factor

- :class:`Unit` - Base class for a unit

- :class:`Units` - Dictionary of units with dynamic prefixing

- :class:`UnitsModule` - Class that wraps a :class:`Units` dictionary as a
  module

.. warning::  Be careful if you use this package to produce conversion factors.
   The factor may be the reciprocal of what you first think.  By the phrase
   "convert from x to y" (where x and y are units), we mean "convert from an
   expression of the quantity in x to an expression of the quantity in y."
   "Quantity in unit" is typically equivalent to "quantity *divided by* unit",
   and to get from *Q*/x (where *Q* is a quantity) to *Q*/y, we must multiply by
   x/y.  For example, the conversion factor from feet to metres is ``ft/m``
   (0.3048) using :mod:`natu.units`.

   - Adapted from SciMath
     (http://docs.enthought.com/scimath/units/intro.html, accessed
     7/10/2014)

.. note::  Known issues:

   - Python's built-in :func:`sum` will only accept dimensionless quantities.
     Consider using :func:`natu.math.fsum` instead.
"""

# pylint: disable=I0011, C0103, C0111, C0301, C0325, E0213, E1101, F0401, R0903
# pylint: disable=I0011, W0142, W0212, W0223, W0621

# Always use floating division:
from __future__ import division

# So that math is the modules from the Python Standard Library:
from __future__ import absolute_import

__author__ = "Kevin Davies"
__email__ = "kdavies4@gmail.com"
__copyright__ = ("Copyright 2013-2014, Kevin Davies, Hawaii Natural Energy "
                 "Institute, and Georgia Tech Research Corporation")
__license__ = "BSD-compatible (see LICENSE.txt)"

__all__ = ('CoherentRelations CompoundDimension CompoundUnit DimObject '
           'Quantity Unit ScalarUnit LambdaUnit Units UnitsModule'.split())

import math
import re

from os.path import dirname
from types import ModuleType
from functools import wraps
# from warnings import warn
from .util import format_e, str2super
from ._prefixes import PREFIXES
from .config import simplification_level, use_quantities, unit_replacements
from .exponents import Exponents, split_code, u, i

try:
    from configparser import (RawConfigParser, ParsingError,
                              Error as ConfigParserError)
except ImportError:
    # For Python 2:
    from ConfigParser import (RawConfigParser, ParsingError,
                              Error as ConfigParserError)

# Compile the formatted unit replacements.
UNIT_REPLACEMENTS = {fmt:
                     [(re.compile(rpl[0]), rpl[1]) for rpl in rpls]
                     for fmt, rpls in unit_replacements.items()}


# Standard functions
# ------------------

def assert_homogeneous(*args):
    r"""Assert that *\*args* have the same dimension.

    If any argument does not have a :attr:`~DimObject.dimension` property, it is
    assumed to be dimensionless.

    **Example:**

    >>> from natu.units import m, ft, s
    >>> assert_homogeneous(m, ft)
    >>> assert_homogeneous(m, s) # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ...
    AssertionError: The quantities must have the same dimension.
    """
    return # TODO: Remove this line and fix the code below.
    try:
        for row in zip(args):
            assert_homogeneous(row) # Recursive
            return
    except TypeError:
        dim = dimension(args[0])
        for arg in args[1:]:
            assert dimension(arg) == dim, \
                "The quantities must have the same dimension."

def dimension(quantity):
    """Return the dimension of *quantity*.

    If *quantity* does not have a :attr:`~DimObject.dimension` property, it is
    assumed to be dimensionless and an empty :class:`CompoundDimension` is
    returned.

    **Example:**

    >>> from natu.units import Pa
    >>> print(dimension(Pa))
    M/(L*T2)
    """
    try:
        return quantity.dimension
    except AttributeError:
        return CompoundDimension()


def dimensionless_value(x):
    """Return the value of the quantity if it is dimensionless.

    Otherwise, raise a :class:`TypeError`.
    """
    try:
        if x.dimensionless:
            return x._value
    except AttributeError:
        return x
    raise TypeError("The quantity is not dimensionless.")


def display(quantity):
    """Return the display unit of *quantity*.

    If *quantity* does not have a :attr:`~DimObject.display` property, then an
    empty :class:`CompoundUnit` is returned.

    **Example:**

    >>> from natu.units import m
    >>> print(display(100*m))
    m
    """
    try:
        return quantity.display
    except AttributeError:
        return CompoundUnit()

def merge(value, prototype):
    """Merge *value* into a new :class:`Quantity` with the properties
    (:attr:`~Quantity.dimension` and :attr:`~Quantity.display`) of *prototype*.

    If *prototype* is not a :class:`~natu.core.Quantity`, then *value* is
    returned directly---as a quantity.

    **Example:**

    >>> from natu.units import m, km
    >>> from natu.core import merge
    >>> merge(value(1000*m), km)
    Quantity(1000, 'L', 'km') (1.0 km)
    """
    try:
        return Quantity(value, prototype.dimension, prototype.display)
    except AttributeError:
        return value


def prohibited(self, other):
    """Not allowed; raises a TypeError"""
    # pylint: disable=I0011, W0613
    raise TypeError("In-pace assignments aren't allowed for units.")


def quantity_only(operation):
    """Return a method that raises a :class:`TypeError` with a message pertinent
    to :class:`LambdaUnit`.
    """

    def prohibited(self, other):
        """Not allowed; raises a TypeError"""
        # pylint: disable=W0613
        raise TypeError(operation.capitalize() + " is only allowed for "
                        "quantities.  Use the lambda unit to generate a "
                        "quantity first.")
    return prohibited

def _times(code):
    """Return a string representing multiplication, depending on the format
    code.
    """
    if code == 'H':
        return '&nbsp;'
    if code == 'L':
        return '\,'
    return ' '

def value(x):
    """Return the :attr:`_value` attribute of *x* if it exists.

    Otherwise, return *x* directly (pass-through).

    **Example:**

    >>> from natu.units import m
    >>> value(10*m) # doctest: +SKIP
    10

    .. testcleanup::
       >>> assert abs(value(10*m) - 10) < 1e-10
    """
    try:
        return x._value
    except AttributeError:
        return x

# Decorators
# ----------


def add_unit(meth):
    """Decorate a method to add a unit to the formatted representation.
    """
    @wraps(meth)
    def wrapped(self, code):

        # Get the display unit.
        self.display = Quantity.unitspace.simplify(self.display)
        unit = Quantity.unitspace(**self.display)
        unit_dim = dimension(unit)
        assert self.dimension == unit_dim, (
            "The display unit ({0.display}; dimension {1}) and the quantity "
            "(dimension {0.dimension}) are incompatible.").format(self,
                                                                  unit_dim)

        # Parse the format code.
        number_code, unit_code = split_code(code)

        # Create the unit string.
        unit_str = format(self.display, unit_code)
        try:
            for rpl in UNIT_REPLACEMENTS[unit_code]:
                unit_str = rpl[0].sub(rpl[1], unit_str)
        except KeyError:
            pass

        return meth(self / unit, number_code, unit_code) + unit_str

    return wrapped


def as_scalarunit(meth):
    """Decorate *meth* so that it returns a :class:`ScalarUnit` if both
    arguments are :class:`ScalarUnit` instances and the result is not
    dimensionless.
    """
    @wraps(meth)
    def wrapped(self, other):
        result = meth(self, other)
        if isinstance(other, ScalarUnit) and isinstance(result, Quantity):
            return ScalarUnit(result._value, result._dimension, result._display)
        return result

    return wrapped


def copy_props(func):
    """Decorate a function to return a :class:`~natu.core.Quantity` that matches
    the first argument except for the computed value.
    """
    @wraps(func)
    def wrapped(x, y):
        return merge(func(x, y), x)

    wrapped.__doc__ += (
        "\nThe :attr:`dimension` and :attr:`display` properties of the first "
        "\nterm are propagated to the result.\n")
    return wrapped


def homogeneous(func):
    """Decorate a function to accept quantities of the same dimension.
    """
    @wraps(func)
    def wrapped(x, y):
        assert_homogeneous(x, y)
        return func(value(x), value(y))

    wrapped.__doc__ += (
        "\nThe terms must have the same dimension.  A quantity can only be "
        "\nadded to, compared to, or subtracted from a non-quantity if the "
        "\nquantity is dimensionless.\n")
    return wrapped


# Classes
# -------

class DefinitionError(Exception):

    """Error in the definition of a unit or constant in an INI file
    """
    pass

class CompoundDimension(Exponents):

    """Dictionary that contains base physical dimensions as keys and exponents
    of those dimensions as values

    This is :class:`natu.exponents.Exponents`, only renamed for clarity.
    Please see that class for details and additional examples.

    **Example:**

    >>> dim = CompoundDimension('L/T')*2
    >>> dim
    CompoundDimension({'L': 2, 'T': -2})
    >>> print(dim)
    L2/T2
    """
    # pylint: disable=I0011, R0904
    pass


class CompoundUnit(Exponents):

    """Dictionary that contains the base units of a display unit as keys and
    exponents of those units as values

    This is :class:`natu.exponents.Exponents`, only renamed for clarity.
    Please see that class for details and additional examples.

    **Example:**

    >>> unit = CompoundUnit('m/s')*2
    >>> unit
    CompoundUnit({'m': 2, 's': -2})
    >>> print(unit)
    m2/s2
    """
    # pylint: disable=I0011, R0904
    pass

# Note that in the DimObject below, dimension and display are properties that
# return copies of the internal _dimension and _display attributes.  Generally,
# only dimension and display should be accessed from the outside to prevent
# mutating the internal _dimension and _display dictionaries.  However, in the
# code below, this rule is strategically broken to avoid the overhead of making
# copies.


class DimObject(object):

    """Base class that records physical dimension and display unit

    **Initialization parameters:**

    - *dimension*: Physical dimension

         This can be a :class:`CompoundDimension` instance, a :class:`dict` of
         similar form, or a string accepted by the
         :meth:`~natu.exponents.Exponents.fromstr` constructor.

    - *display*: Display unit

         This can be a :class:`CompoundUnit` instance, a :class:`dict` of
         similar form, or a string accepted by the
         :meth:`~natu.exponents.Exponents.fromstr` constructor.

    Here, the physical dimension and the display unit are not checked for
    consistency.

    :attr:`dimension` and :attr:`dimensionless` are read-only attributes (see
    below), but :attr:`display` can be set using the same format as the
    *display* argument above.
    """

    def __init__(self, dimension, display):
        """Initialize by setting the physical dimension and display unit.

        See the top-level class documentation.
        """
        self._dimension = CompoundDimension(dimension)
        self._display = CompoundUnit(display)

    @property
    def dimension(self):
        """Physical dimension as a :class:`CompoundDimension` instance"""
        return self._dimension.copy()

    @property
    def dimensionless(self):
        """*True* if the instance is dimensionless"""
        return not self._dimension

    @property
    def display(self):
        """Display unit as a :class:`CompoundUnit` instance"""
        return self._display.copy()

    @display.setter
    def display(self, display):
        """Set the display unit.

        *display* can be a :class:`CompoundUnit` instance, a :class:`dict` of
        similar form, or a string accepted by the
        :meth:`~natu.exponents.Exponents.fromstr` constructor.

        Here, the display unit is not checked for dimensional consistency (with
        :attr:`dimension`).
        """
        self._display = CompoundUnit(display)


class Quantity(DimObject):

    """Class to represent a physical quantity

    **Initialization parameters:**

    - *value*: Value of the quantity (a :class:`numbers.Number` instance)

         This be expressed as the product of a number and the value of a
         unit.  It is independent of the unit since the number scales inversely
         to the unit.

    - *dimension*: Physical dimension

         This can be a :class:`CompoundDimension` instance, a :class:`dict` of
         similar form, or a string accepted by the
         :meth:`~natu.exponents.Exponents.fromstr` constructor.

    - *display*: Display unit

         This can be a :class:`CompoundUnit` instance, a :class:`dict` of
         similar form, or a string accepted by the
         :meth:`~natu.exponents.Exponents.fromstr` constructor.

    **Examples:**

    Instantiation and simple operations:

    >>> import natu.units
    >>> mass = Quantity(1, 'M', 'kg')
    >>> velocity = Quantity(1, 'L/T', 'm/s')
    >>> energy = mass*velocity**2
    >>> print(energy.dimension)
    L2*M/T2
    >>> print(energy.display)
    kg*m2/s2

    However, it is easier to create a quantity as the product of a number and
    unit(s); see :mod:`natu.units`.

    Changing the display unit:

    >>> mass.display = 'lb'
    >>> print(mass) # doctest: +ELLIPSIS
    2.20462... lb

    Note that the value has not changed:

    >>> mass # doctest: +ELLIPSIS
    Quantity(1, 'M', 'lb') (2.20462... lb)

    Although the :attr:`display` property can be changed, quantities are
    otherwise immutable.  The in-place operators create new instances:

    >>> initial_id = id(velocity)
    >>> velocity *= 0.5
    >>> id(velocity) == initial_id
    False


    .. _Python: https://www.python.org/
    """

    unitspace = None
    """Underlying unit dictionary used to define quantities"""

    def __init__(self, value, dimension, display=''):
        """Initialize a quantity by setting the value, physical dimension, and
        display unit.

        See the top-level class documentation.
        """
        # Set the value.
        self._value = value

        # Set the dimension and display unit.
        DimObject.__init__(self, dimension, display)

    @copy_props
    @homogeneous
    def __add__(x, y):
        """x.__add__(y) <==> x + y
        """
        return x + y

    def __radd__(x, y):
        """x.__radd__(y) <==> y + x
        """
        return y + dimensionless_value(x)

    @copy_props
    @homogeneous
    def __sub__(x, y):
        """x.__sub__(y) <==> x - y
        """
        return x - y

    def __rsub__(x, y):
        """x.__add__(y) <==> y - x
        """
        return y - dimensionless_value(x)

    def __neg__(x):
        """x.__neg__() <==> -x
        """
        return merge(-x._value, x)

    def __pos__(x):
        """x.__pos__() <==> x
        """
        return x

    def __mul__(x, y):
        """x.__mul__(y) <==> x*y
        """
        try:
            value = y._value * x._value # Product of quantities
        except AttributeError:
            if isinstance(y, LambdaUnit):
                return NotImplemented  # Defer to LambdaUnit's _toquantity().
            return Quantity(y * x._value, x.dimension, x.display)
        dimension = y._dimension + x._dimension
        if dimension:
            return Quantity(value, dimension, y._display + x._display)
        return value

    __rmul__ = __mul__

    def __truediv__(x, y):
        """x.__truediv__(y) <==> x/y
        """
        try:
            value = x._value / y._value
        except AttributeError:
            if isinstance(y, DimObject):
                return NotImplemented  # Defer; y could be a LambdaUnit.
            return Quantity(x._value / y, x.dimension, x.display)
        dimension = x._dimension - y._dimension
        if dimension:
            return Quantity(value, dimension, x._display - y._display)
        return value

    __div__ = __truediv__

    def __rtruediv__(x, y):
        """x.__rtruediv__(y) <==> y/x
        """
        try:
            value = y._value / x._value
        except AttributeError:
            return Quantity(y / x._value, -x._dimension, -x._display)
        dimension = y._dimension - x._dimension
        if dimension:
            return Quantity(value, dimension, y._display - x._display)
        return value

    __rdiv__ = __rtruediv__

    def __pow__(x, y):
        """x.__pow__(y) <==> pow(x, y)

        Only the 2-argument version of :func:`pow` is supported since the value
        of a quantity is not generally an integer.
        """
        if y == 0:
            return 1
        try:
            y = dimensionless_value(y)
        except TypeError:
            raise TypeError("The exponent must be dimensionless.")
        except AttributeError:
            pass
        return x.__class__(x._value ** y, x._dimension * y, x._display * y)

    def __rpow__(x, y):
        """y.__pow__(x) <==> pow(x, y)

        This is only implemented for dimensionless quantities.

        Only the two-argument version of :func:`pow` is supported since the
        value of a quantity is not generally an integer.
        """
        if x.dimensionless:
            return y ** x._value
        raise TypeError("The exponent must be dimensionless.")

    @add_unit
    def __format__(number, number_code, unit_code):
        """Format the quantity as a string according to *code*.
        """
        number_str = format_e(format(number, number_code), unit_code)
        return number_str + _times(unit_code)

    def __repr__(self):
        """Return a string representing the quantity as the product of a number
        and a unit.

        **Example:**

        >>> from natu.units import *
        >>> print(1*m)
        1.0 m
        """
        return format(self)

    def __int__(self):
        """Return the quantity as an integer.

        This is only possible if the quantity is dimensionless.
        """
        return int(dimensionless_value(self))

    def __float__(self):
        """Return the quantity as a float.

        This is only possible if the quantity is dimensionless.
        """
        return float(dimensionless_value(self))

    def __trunc__(self):
        """Return the :class:`numbers.Real` value *x* truncated to an
        :class:`numbers.Integral` (usually a long integer).

        This method is used by :func:`natu.math.trunc` and relies on
        :func:`math.trunc`.
        """
        return math.trunc(float(self))

    def __abs__(self):
        """x.__abs__() <==> abs(x)
        """
        return merge(abs(self._value), self)

    def __eq__(x, y):
        """x.__eq__(y) <==> x==y
        To be considered equal, it is not necessary that quantities have the
        same display unit (:attr:`display`)---only the same value and dimension.
        """
        try:
            return (x._dimension == y._dimension and
                    x._value == y._value)
        except AttributeError:
            return x.dimensionless and x._value == y

    def __ne__(x, y):
        """x.__ne__(y) <==> x!=y

        To be considered equal, it is not necessary that quantities have the
        same display unit (:attr:`display`)---only the same value and dimension.
        A quantity is equal to zero if its value is zero, regardless of
        dimension.
        """
        try:
            return (x._dimension != y._dimension or
                    x._value != y._value)
        except AttributeError:
            return not x.dimensionless or x._value != y

    def __bool__(self):
        """self != 0
        """
        return self._value != 0

    __nonzero__ = __bool__  # For Python2

    @homogeneous
    def __ge__(x, y):
        """x.__ge__(y) <==> x>=y
        """
        return x >= y

    @homogeneous
    def __le__(x, y):
        """x.__le__(y) <==> x<=y
        """
        return x <= y

    @homogeneous
    def __gt__(x, y):
        """x.__gt__(y) <==> x>y
        """
        return x > y

    @homogeneous
    def __lt__(x, y):
        """x.__lt__(y) <==> x<y
        """
        return x < y

    @homogeneous
    def __divmod__(x, y):
        """x.__divmod__(y) <==> divmod(x, y)
        """
        return x // y, x % y

    @homogeneous
    def __rdivmod__(x, y):
        """x.__rdivmod__(y) <==> divmod(y, x)
        """
        return y // x, y % x

    @homogeneous
    def __floordiv__(x, y):
        """x.__floordiv__(y) <==> x//y
        """
        return x // y

    @homogeneous
    def __rfloordiv__(x, y):
        """x.__rfloordiv__(y) <==> y//x
        """
        return y // x

    @homogeneous
    def __mod__(x, y):
        """x.__mod__(y) <==> x%y
        """
        return x % y

    @homogeneous
    def __rmod__(x, y):
        """x.__mod__(y) <==> y%x
        """
        return y % x


class Unit(DimObject):

    """Base class for a unit

    In-place operations are blocked because the value of a unit is predefined
    and should not be changed.

    **Initialization parameters:**

    - *dimension*: Physical dimension

         This can be a :class:`CompoundDimension` instance, a :class:`dict` of
         similar form, or a string accepted by the
         :meth:`~natu.exponents.Exponents.fromstr` constructor.

    - *display*: Display unit

         This can be a :class:`CompoundUnit` instance, a :class:`dict` of
         similar form, or a string accepted by the
         :meth:`~natu.exponents.Exponents.fromstr` constructor.

    - *prefixable*: *True* if the unit can be prefixed

    :attr:`~DimObject.dimension`, :attr:`~DimObject.dimensionless`, and
    :attr:`prefixable` are read-only attributes, but :attr:`~DimObject.display`
    can be set using the same format as the *display* argument above.
    """

    def __init__(self, dimension, display, prefixable=False):
        """Initialize by setting the dimension, display unit, and prefixable
        flag.

        See the top-level class documentation.
        """
        DimObject.__init__(self, dimension, display)
        self._prefixable = prefixable

    @property
    def prefixable(self):
        """*True* if the unit can be prefixed

        **Example:**

        >>> from natu.units import m, ft
        >>> m.prefixable
        True
        >>> ft.prefixable
        False
        """
        return self._prefixable

    # No in-place operations
    __iadd__ = prohibited
    __idiv__ = prohibited
    __ifloordiv__ = prohibited
    __imul__ = prohibited
    __ipow__ = prohibited
    __isub__ = prohibited
    __itruediv__ = prohibited


class ScalarUnit(Quantity, Unit):

    """Unit that involves just a scaling factor

    **Initialization parameters:**

    - *value*: Value of the unit (a :class:`numbers.Number` instance)

    - *dimension*: Physical dimension

         This can be a :class:`CompoundDimension` instance, a :class:`dict` of
         similar form, or a string accepted by the
         :meth:`~natu.exponents.Exponents.fromstr` constructor.

    - *display*: Display unit

         This can be a :class:`CompoundUnit` instance, a :class:`dict` of
         similar form, or a string accepted by the
         :meth:`~natu.exponents.Exponents.fromstr` constructor.

    - *prefixable*: *True* if the unit can be prefixed

    **Examples:**

    >>> from natu.units import kg, m, s
    >>> # F
    >>> kg
    ScalarUnit(1, 'M', 'kg', False) (kg)
    >>> kg*m**2/s**2
    ScalarUnit(1, 'L2*M/T2', 'J', False) (J)

    Note that the display unit can be changed:

    >>> m.display = 'ft'
    >>> m # doctest: +ELLIPSIS
    ScalarUnit(1, 'L', 'ft', True) (3.2808... ft)

    .. testcleanup::
       >>> m.display = 'm'

    Now any quantity generated from the metre (m) will display in feet (ft)
    instead. However, the value is unchanged; the metre still represents the
    same length.
    """

    def __init__(self, value, dimension, display='', prefixable=False):
        """Initialize a scalar unit by setting the value, physical dimension,
        display unit, and prefixable flag.

        See the top-level class documentation.
        """
        # pylint: disable=I0011, W0231

        # Set the value.
        self._value = value

        # Set the dimension, display unit, and prefixable flag.
        Unit.__init__(self, dimension, display, prefixable)

    @classmethod
    def from_quantity(cls, quantity, display, prefixable=False):
        """Convert a quantity (instance of :class:`Quantity`) to a scalar unit.

        The value and dimension are taken from *quantity*.  The display unit
        must be provided (via *display*).

        **Example:**

        >>> from natu.units import ns
        >>> shake = ScalarUnit.from_quantity(10*ns, 'shake')

        .. testcleanup::
           >>> from natu import units
           >>> units.shake = shake
           >>> shake
           ScalarUnit(1e-08, 'T', 'shake', False) (shake)
        """
        quantity.__class__ = cls
        quantity.display = display
        quantity._prefixable = prefixable
        return quantity

    def __repr__(self):
        """Return a string representation of the scalar unit.

        The first part (``ScalarUnit(...)``) is the expression that would
        generated the unit.  The part in the last parentheses is the value as
        the product of a number and a unit.
        """
        # Run this first to simplify self.display (see Units.load_ini):
        desc = "ScalarUnit %s" % self._display
        desc = ("dimensionless {}" if self.dimensionless else
                "{} with dimension %s" % self._dimension).format(desc)
        desc += " (prefixable)" if self._prefixable else " (not prefixable)"
        return desc

    @as_scalarunit
    def __mul__(x, y):
        """x.__mul__(y) <==> x*y
        """
        return Quantity.__mul__(x, y)

    __rmul__ = __mul__

    @as_scalarunit
    def __truediv__(x, y):
        """x.__truediv__(y) <==> x/y
        """
        return Quantity.__truediv__(x, y)

    __div__ = __truediv__

    @as_scalarunit
    def __rtruediv__(x, y):
        """x.__truediv__(y) <==> x/y
        """
        return Quantity.__rtruediv__(x, y)

    __rdiv__ = __rtruediv__

    @add_unit
    def __format__(number, number_code, unit_code):
        """Format the scalar unit as a string according to *code*.
        """

        TOLERANCE = 1e-14
        if abs(number - 1) < TOLERANCE:
            # Exclude the number since it's nearly 1.
            return ''
        else:
            number_str = format_e(format(number, number_code), unit_code)
            return number_str + _times(unit_code)


class LambdaUnit(Unit):

    """Unit that involves a offset or other operations besides scaling

    **Initialization parameters:**

    - *toquantity*: Function that maps a number to a quantity

    - *tonumber*: Function that maps a quantity to a number

    - *dimension*: Physical dimension

         This can be a :class:`CompoundDimension` instance, a :class:`dict` of
         similar form, or a string accepted by the
         :meth:`~natu.exponents.Exponents.fromstr` constructor.

    - *display*: Display unit

         This can be a :class:`CompoundUnit` instance, a :class:`dict` of
         similar form, or a string accepted by the
         :meth:`~natu.exponents.Exponents.fromstr` constructor.

    - *prefixable*: *True* if the unit can be prefixed

    **Examples:**

    >>> from natu.units import degC, K
    >>> degC # doctest: +ELLIPSIS
    LambdaUnit(...)
    >>> 25*degC/K
    298.15
    """

    def __init__(self, toquantity, tonumber, dimension, display='',
                 prefixable=False):
        """Initialize a lambda unit by setting the function and its inverse,
        physical dimension, display unit, and prefixable flag.

        See the top-level class documentation.
        """
        # pylint: disable=I0011, R0913
        # Set the methods that map a number to a quantity and vice versa.
        self._toquantity = toquantity
        self._tonumber = tonumber

        # Set the dimension, display unit, and prefixable flag.
        Unit.__init__(self, dimension, display, prefixable)

    def __repr__(self):
        """Return a string represention of the lambda unit.
        """
        desc = "LambdaUnit %s" % self._display
        desc = ("dimensionless {}" if self.dimensionless else
                "{} with dimension %s" % self._dimension).format(desc)
        desc += " (prefixable)" if self._prefixable else " (not prefixable)"
        return desc

    def __pow__(x, y):
        """x.__pow__(y) <==> pow(x, y)

        Only the 2-argument version of :func:`pow` is supported.  In addition, y
        must be 1, 0, or -1.  With y=1, the result is the lambda unit itself.
        With y=0, the result is 1 (unity).  With y=-1, the result is the inverse
        of the lambda unit.
        """
        if y == 1:
            return x
        if y == 0:
            return 1
        if y == -1:
            return LambdaUnit(x._tonumber, x._toquantity,  # Swapped
                              -x._dimension, -x._display)
        raise TypeError("Lambda units can only be raised to the power of -1, "
                        "0, or 1.")

    def __rmul__(unit, number):
        """unit.__rmul__(number) <==> number*unit
        """
        display = unit.display
        if isinstance(number, Quantity):
            assert not isinstance(number, Unit), (
                "Lambda units cannot be combined with other units.")
            assert number.dimensionless, (
                "The argument to the lambda unit must be dimensionless.")
            display += number.display
            as_quantity = False
        else:
            as_quantity = bool(display) and use_quantities
        try:
            quantity = unit._toquantity(number)
        except AssertionError:
            raise AssertionError("The number isn't a valid argument for the "
                                 "lambda unit.")
        if isinstance(quantity, Quantity):
            quantity.display = display
            return quantity
        elif as_quantity:
            return Quantity(quantity, {}, display)
        return quantity

    def __rtruediv__(unit, quantity):
        """unit.__rtruediv__(quantity) <==> quantity/unit
        """
        display = -unit._display
        if isinstance(quantity, Quantity):
            display += quantity.display
            as_quantity = False
        else:
            as_quantity = display and unit.dimensionless and use_quantities
        try:
            number = unit._tonumber(quantity)
        except AssertionError:
            raise AssertionError("The quantity isn't a valid argument for the "
                                 "inverse of the lambda unit.")
        if isinstance(number, Quantity):
            assert number.dimensionless, ("The result of the inverse of the "
                                          "lambda unit should be "
                                          "dimensionless.")
            number.display = display
            return number
        elif as_quantity:
            # E.g., keep unit of B-1 when doing 1/B but not unit of K/degC when
            # doing 300*K/degC.
            return Quantity(number, {}, display)
        return number

    __rdiv__ = __rtruediv__

    # Provide helpful messages for the unsupported methods (that are supported
    # by Quantity).
    def __mul__(self, other):
        """Not supported
        """
        # pylint: disable=I0011, R0201
        raise AttributeError("Lambda units can only be used on the right side "
                             "of a product.")

    def __truediv__(self, other):
        """Not supported
        """
        # pylint: disable=R0201, W0613
        raise AttributeError("Lambda units can only be used as the denominator "
                             "of a quotient.")

    __div__ = __truediv__

    # Addition, etc.
    __add__ = quantity_only('Addition')
    __radd__ = quantity_only('Addition')
    __pos__ = quantity_only('The positive operator')
    # Subtraction, etc.
    __sub__ = quantity_only('Subtraction')
    __rsub__ = quantity_only('Subtraction')
    __neg__ = quantity_only('Negation')
    # Exponentiation
    __rpow__ = quantity_only('Exponentiation')
    # Modulo
    __mod__ = quantity_only('The modulo operator')
    __rmod__ = quantity_only('The modulo operator')
    # Division modulo
    __divmod__ = quantity_only('Division modulo')
    __rdivmod__ = quantity_only('Division modulo')
    # Floor division
    __floordiv__ = quantity_only('Floor division')
    __rfloordiv__ = quantity_only('Floor division')
    # Comparison
    __cmp__ = quantity_only('Comparison')
    __ge__ = quantity_only('Comparison')
    __gt__ = quantity_only('Comparison')
    __le__ = quantity_only('Comparison')
    __lt__ = quantity_only('Comparison')
    __nonzero__ = quantity_only('Comparison')
    # Other
    __abs__ = quantity_only('Absolute value')


class CoherentRelations(list):

    """List of coherent relations among units

    Each item is a tuple consisting of a relation and a set of the dimensions
    involved in the units of the relation.  A relation is expressed as a
    :class:`CompoundUnit` instance that evaluates to unity.
    """
    pass


class Units(dict):

    """Dictionary of units with dynamic prefixing (upon access)

    **Properties:**

    - :attr:`coherent_relations` - List of coherent relations among the units

         Each entry is a tuple of a :class:`CompoundUnit` instance that
         evaluates to unity and a set of the dimensions involved in the units in
         the :class:`CompoundUnit`.
    """

    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

        # Initialize an empty list of coherent relations.
        self.coherent_relations = CoherentRelations()

    def __getitem__(self, symbol):
        """Access a unit by *symbol* (a string).

        Prefixes are supported.

        **Example:**

        >>> from natu.units import _units
        >>> _units['psi']
        ScalarUnit(6894.76, 'M/(L*T2)', 'psi', False) (psi)
        """
        try:
            return dict.__getitem__(self, symbol)  # Constant or standard unit
        except KeyError:
            pass  # Prefixed unit (below)

        # Default error:
        error = KeyError(symbol + " isn't a valid unit.")

        # Try prefixes.
        for len_prefix in [1, 2]:  # A prefix can have 1 or 2 characters.

            # Get the base unit.
            try:
                basesymbol = symbol[len_prefix:]
                baseunit = dict.__getitem__(self, basesymbol)
            except IndexError:
                # The unit isn't longer than the length of the prefix.
                raise error
            except KeyError:
                # The base symbol isn't a valid unit, so try a longer prefix.
                continue

            # Check that the unit is prefixable.
            try:
                assert(not use_quantities or baseunit.prefixable)
            except (AttributeError, AssertionError):
                error = KeyError(basesymbol + " isn't prefixable.")
                continue # Try a longer prefix.

            # Look up and apply the prefix.
            prefix = symbol[:len_prefix]
            try:
                p = PREFIXES[prefix]
            except KeyError:
                error = KeyError(prefix + " isn't a valid prefix.")
            else:
                if isinstance(baseunit, ScalarUnit):
                    return ScalarUnit(p * baseunit._value,
                                      baseunit.dimension, symbol)
                if isinstance(baseunit, LambdaUnit):
                    return LambdaUnit(lambda n: baseunit._toquantity(p * n),
                                      lambda q: baseunit._tonumber(q) / p,
                                      baseunit.dimension, symbol)
                return p * baseunit # Scalar unit, but not using quantities

        raise error

    def __call__(self, **factors):
        r"""Generate a compound, coherent unit from existing units.

        **Parameters:**

        - *\*\*factors*: Factors of the compound unit

          The factors are expressed as keyword arguments where the keys are unit
          symbols and the values are exponents.

        **Example:**

        >>> from natu.units import _units
        >>> _units(lbf=1, inch=-2)
        ScalarUnit(6894.76, 'M/(L*T2)', 'psi', False) (psi)

        Here, the unit was automatically simplified to psi using
        :attr:`_units.coherent_relations`.
        """
        factors = [self[base] ** exp for base, exp in factors.items()]
        return reduce(lambda x, y: x * y, factors)

    def load_ini(self, files):
        r"""Add units to the unit dictionary from a \*.ini file or list of files
        (*files*).

        **Examples:**

        .. code-block:: python

           >>> from natu.core import Units
           >>> from natu.config import definitions

           >>> units = Units()
           >>> units.load_ini([definitions[0]])
           >>> units.keys() # doctest: +SKIP
           ['R', 'R_K', 'R_inf', 'c', 'k_Aprime', 'k_F', 'k_J', 'rational']

        .. testcleanup::

           >>> sorted(units.keys())
           ['R', 'R_K', 'R_inf', 'c', 'k_Aprime', 'k_F', 'k_J', 'rational']
        """
        # pylint: disable=I0011, R0912, R0914

        # Temporarily add some constants, functions, and classes to the unit
        # space for use in the *.ini files.
        from fractions import Fraction
        sqrt = lambda x: x**Fraction(0.5)
        # (Not using natu.math.sqrt to avoid cyclic import.)
        provided = dict(pi=math.pi, exp=math.exp, log=math.log,
                        log10=math.log10, sqrt=sqrt,
                        Quantity=Quantity, ScalarUnit=ScalarUnit)
        self.update(provided)

        # Load the units from the *.ini files.
        try:
            config = RawConfigParser(interpolation=None,
                                     inline_comment_prefixes=[';'])
        except TypeError:
            config = RawConfigParser()
        config.optionxform = str  # Units are case sensitive.
        if len(config.read(files)) != len(files):
            raise DefinitionError("Failed to open/find all definition files")
        for section in config.sections():
            for symbol, value in config.items(section):
                # print(symbol)
                if symbol in self:
                    msg = ('In section "%s", overriding previous value of %s'
                           % (section, symbol))
                    # warn(msg)
                    print(msg)
                try:
                    unit = eval(value, self, self)
                    # self is provided as the global namespace as well as the
                    # local one so that it's immediately used by the lambda
                    # functions.  This doesn't allow prefixes in the lambda
                    # expressions since eval() doesn't use __getitem__() for
                    # globals.
                    # TODO: Consider using pyparsing instead of eval() (for
                    # safety) if it's not too slow.
                    if isinstance(unit, tuple):
                        unit, prefixable = unit
                        if isinstance(unit, tuple):
                            # The unit is a lambda unit, defined via a tuple.
                            toquantity, tonumber = unit
                            try:
                                # Evaluate the unit with an arbitrary number
                                # (zero) to determine the dimension.
                                dim = toquantity(0).dimension
                            except AttributeError:
                                # The result doesn't have a dimension; the unit
                                # must be dimensionless.
                                dim = {}
                            unit = LambdaUnit(toquantity, tonumber, dim, symbol,
                                              prefixable)
                        elif isinstance(unit, LambdaUnit):
                            # The unit is a lambda unit, defined directly.
                            unit = LambdaUnit(unit._toquantity, unit._tonumber,
                                              unit._dimension, unit._display,
                                              prefixable)
                        elif isinstance(unit, Quantity):
                            # The unit is a scalar unit with dimension.
                            if (isinstance(unit, ScalarUnit)
                                and 'ScalarUnit' not in value):
                                # The unit has been coherently derived.
                                dimensions = set(unit.dimension)
                                for u in unit.display.keys():
                                    dimensions = dimensions.union(
                                        self[u].dimension)
                                relation = (unit.display - {symbol: 1},
                                            dimensions)
                                self.coherent_relations.append(relation)
                            unit = ScalarUnit.from_quantity(unit, symbol,
                                                           prefixable)
                        else:
                            # The unit is a dimensionless scalar unit.
                            unit = ScalarUnit(unit, {}, symbol, prefixable)
                except (AssertionError, AttributeError, ConfigParserError,
                        NameError, SyntaxError, TypeError, ValueError) as e:
                    raise DefinitionError("can't load '%s' due to %s"
                                          % (symbol, type(e).__name__))
                if isinstance(unit, Quantity) and not use_quantities:
                    # Represent quantities as pure numbers (don't track the
                    # dimension and display unit).
                    unit = unit._value
                self[symbol] = unit

        # Remove the temporary items.
        for key in provided.keys():
            del self[key]
        # __builtins__ got added because self is the global argument to eval()
        # above.
        self.pop('__builtins__', None)

    def simplify(self, unit, level=simplification_level):
        r"""Simplify a compound unit.

        This function seeks to minimize the sum of the absolute values of the
        exponents of the base factors by substituting coherently related units.
        It does not use units that involve dimensions not included in the
        original representation.  It uses the internal
        :attr:`coherent_relations` list which is generated while parsing the
        \*.ini files (in :meth:`load_ini`).  It will not always find the
        simplest representation because some simplifications involve first
        making the representation more complex.

        **Parameters:**

        - *unit*: Unit to be simplified

             This can be a :class:`CompoundUnit` instance, a :class:`dict` of
             similar form, or a string accepted by the
             :meth:`~natu.exponents.Exponents.fromstr` constructor.

        - *level*: Number of levels of recursion to perform

             This is the number of non-minimizing substitutions that can be made
             while seeking the simplest representation of the unit.

        **Returns:**  A :class:`CompoundUnit` instance indicating the new
        representation of the unit.

        **Example:**

        >>> # High-level/indirect:
        >>> from natu.units import *
        >>> print(kg*m**2/s**2)
        J

        >>> # Low-level/direct:
        >>> from natu.units import _units
        >>> print(_units.simplify('kg*m2/s2'))
        J
        """
        # pylint: disable=I0011, E1103

        # Objective to minimize
        complexity = lambda x: sum(abs(exp) for exp in x.values())
        # This is the L1 norm (sum of the absolute values of the exponents),
        # making this problem L1 minimization.  There isn't a simple solution,
        # and there isn't a simple Python package to find it without a lot of
        # dependencies (as of 6/29/14).  The closest packages are scipy.linalg,
        # L1L2Py, and pyl1min (http://sourceforge.net/projects/pyl1min/).  The
        # approach below is more or less brute force.  It isn't guaranteed to
        # find the best solution, but it's straightforward to implement and
        # works well enough.

        # In case the unit is a string:
        unit = CompoundUnit(unit)

        # Shortcut---no simplication:
        if level == 0 or complexity(unit) <= 1:
            return unit

        # Set of dimensions involved in the unit:
        dimensions = set(self(**unit).dimension)

        # Loop to try each of the coherent relations that involve a subset of
        # the dimensions of the unit.
        simpler = True
        while simpler:
            simpler = False
            for identity, dim in self.coherent_relations:
                # if not dim.issubset(dimensions):
                    # Skip; the relation involves different dimensions.
                #     continue
                common = set(identity).intersection(unit)
                if len(common) < len(identity) / 2 - 0.5:
                    # Skip for speed; the relation isn't worth it.
                    continue
                for factor in common:
                    factor = float(unit[factor]) / identity[factor]
                    int_factor = int(factor)
                    if int_factor == factor:
                        temp = unit - identity * int_factor
                        if level > 1:
                            temp = self.simplify(temp, level - 1) # Recursion
                        if complexity(temp) < complexity(unit):
                            unit = temp
                            simpler = True
                            break
        return unit


class UnitsModule(ModuleType):

    r"""Class that wraps a :class:`Units` dictionary as a module

    **Initialization parameters:**

    - *module*: The module into which the units should be wrapped

         This must have the attributes :attr:`__name__` and :attr:`__doc__`.

    - *definitions*: A \*.ini file, list of files, or dictionary with
      units that should be inserted into the module

         It is not necessary to provide prefixed versions of the units unless
         they should be available via wildcard import (e.g.,
         ``from units_module import *``).

    Only one :class:`UnitsModule` can be instantiated directly from \*.ini files
    per Python_ session.  This prevents conflicts that could arise if the units
    were reloaded with different base constants.
    """

    def __init__(self, module, definitions):
        r"""Initialize the module with meta attributes matching those of
        *module* and units and constants from a \*.ini file or list of files
        (*definitions*).
        """
        # pylint: disable=I0011, E1002, W0233

        # Copy attributes from the original module.
        super(UnitsModule, self).__init__(module.__name__, module.__doc__)
        for attr in ['__author__', '__email__', '__copyright__', '__license__',
                     '__file__', '__package__']:
            try:
                setattr(self, attr, getattr(module, attr))
            except AttributeError:
                pass
        self.__path__ = dirname(self.__name__)  # Necessary in Python3; why?

        # Load the units.
        if isinstance(definitions, dict):
            # Create a unit dictionary from the provided dictionary.
            self._units = Units(definitions)
        else:
            # Create an empty unit dictionary.
            self._units = Units()

            # Monkey-patch the Quantity class to use only that unit dictionary.
            assert not Quantity.unitspace, (
                "The units module can only be loaded once.")
            Quantity.unitspace = self._units

            # Load units from the ini files.
            try:
                self._units.load_ini(definitions)
            except (DefinitionError, ParsingError):
                # Allow the user to fix the INI files and try to import again.
                Quantity.unitspace = None
                raise

        self.__all__ = list(self._units)

    def __getattr__(self, name):
        """Return a unit or attribute matching *name* (a string).

        Prefixes are accepted on prefixable units.
        """
        try:
            return self._units[name]
        except KeyError:
            try:
                return getattr(self._units, name)
            except AttributeError:
                raise AttributeError(name + " isn't a valid unit or attribute")

    def __setattr__(self, name, value):
        """Set an attribute (*name*) to a value (*value*).

        If *value* is a :class:`DimObject`, it is added to the internal units
        dictionary (:attr:`_units`).  Otherwise, it is added as a standard
        attribute.
        """
        if isinstance(value, DimObject):
            self._units[name] = value
        else:
            ModuleType.__setattr__(self, name, value)



#!/usr/bin/python
"""Aliases for constants and units of length, with support for prefixes
"""
# pylint: disable=I0011, E0611

from natu.groups import _update_module
from ..units import _units, m

_update_module(__name__, _units, m.dimension)



#!/usr/bin/python
"""`Python math`_, adapted for use with physical quantities

Many of the functions only accept dimensionless quantities, and they operate on
the underlying values---not the values as represented in a particular display
unit.  To operate on values in a display unit would be to favor a particular
unit, which is against the `design of the package <index.html>`_.

The constants (*pi*, *e*) are exactly as they are in `Python math`_.

These functions accept floats, integers, and dimensionless quantities:

- Number-theoretic and representation functions: :func:`factorial`,
  :func:`frexp`, :func:`modf`, and :func:`trunc`
- Power and logarithmic functions: :func:`exp`, :func:`expm1`, :func:`log10`,
  :func:`log1p`, and :func:`log`
- Hyperbolic functions: :func:`acosh`, :func:`asinh`, :func:`atanh`,
  :func:`cosh`, :func:`sinh`, and :func:`tanh`
- Special functions: :func:`erf`, :func:`erfc`, :func:`gamma`, and
  :func:`lgamma`

They are the same as those from `Python math`_, except the documentation of
:func:`acosh`, :func:`asinh`, and :func:`atanh` has been corrected (see
http://bugs.python.org/issue21902).  The functions cast quantities cast as
floats and return floats, except for :func:`factorial`, which casts quantities
as integers and returns integers.

All of the other functions (below) are different than those in `Python math`_.

These functions accept angle as a quantity:

- :func:`cos`, :func:`sin`, and :func:`tan`

They return floats (as in `Python math`_).

These functions accept floats, integers, and dimensionless quantities:

- :func:`acos`, :func:`asin`, :func:`atan`, and :func:`atan2`

They return angles as quantities.

:func:`atan2` accepts accept floats, integers, and quantities of the same
dimension.  It returns angle as a quantity.

These functions are no longer applicable and have been deleted since angle is a
quantity:

- :func:`degrees` and :func:`radians`

These functions accept floats, integers, and dimensionless quantities:

- :func:`ceil` and :func:`floor`

If the input is a float or an integer, the output is a float.  If the input is
a :class:`~natu.core.Quantity` or a :class:`~natu.core.ScalarUnit`, the result
is the same and has the same :attr:`dimension`, :attr:`display`, etc.

These functions accept floats, integers, and quantities:

- :func:`fabs` and :func:`copysign`

If the input is a float or an integer, the output is a float.  If the input is a
:class:`~natu.core.Quantity` or a :class:`~natu.core.ScalarUnit`, the result is
of the same type and has the same :attr:`dimension`, :attr:`display`, etc. (of
the first argument in the case of :func:`copysign`).

These functions also accept floats, integers, and quantities:

- :func:`ldexp`, :func:`pow`, and :func:`sqrt`

If the input is an float or an integer, the output is a float.  If the input is
a :class:`~natu.core.Quantity` or a :class:`~natu.core.ScalarUnit`, the result
is the same.  The dimensions and display units are handled according to the
power.

:func:`fmod` accepts floats, integers, and quantities of the same dimension.
The output is always a float.

These functions also accept floats, integers, and quantities of the same
dimension:

- :func:`fsum` and :func:`hypot`

The display unit (and :attr:`prefixable` attribute, if applicable) of the first
argument or entry is propagated to the output.

These functions accept floats or quantities:

- :func:`isinf`, :func:`isfinite` (only available in Python >= 3.2), and
  :func:`isnan`

Only the value of a quantity is used; dimension and display unit are ignored.


.. _Python math: https://docs.python.org/3/library/math.html
"""
# pylint: disable=I0011, C0103, E0601, W0401, W0614, W0622

from __future__ import absolute_import

from math import *
from fractions import Fraction
from . import _decorators as decor

# Number-theoretic and representation functions
# ---------------------------------------------
ceil = decor.dimensionless_copy_props(
    ceil,
    """Return the ceiling of x as a float or quantity.

    This is the smallest integral value >= x.""")

floor = decor.dimensionless_copy_props(
    floor,
    """Return the floor of x as a float or quantity.

    This is the largest integral value <= x.""")

copysign = decor.use_values_copy_props(copysign)

fabs = decor.use_value_copy_props(
    fabs,
    "Return the absolute value of the float or quantity x.")

fmod = decor.homogeneous(fmod)

fsum = decor.homogeneous_copy_props_iter(fsum)

try:
    isinfinite = decor.use_value(isinfinite)  # This wasn't defined until v3.2.
except NameError:
    pass

isinf = decor.use_value(
    isinf,
    "Check if float or quantity x is infinite (positive or negative).")

isnan = decor.use_value(
    isnan,
    "Check if float or quantity x is not a number (NaN).")

ldexp = decor.use_values_copy_props_xi(
    ldexp,
    "Return x * (2**i), where x is a float or a quantity.")

# factorial, frexp, modf, and trunc are ok as they are.

# Power and logarithmic functions
# -------------------------------
pow = decor.use_value_raise(pow)
sqrt = decor.use_value_raise(sqrt, Fraction(1, 2))
# exp, expm1, log, log1p, and log10 are ok as they are.

# Trigonometric functions
# -----------------------
acos = decor.inv_trig(acos, "Return the arc cosine of x as an angle.")
asin = decor.inv_trig(asin, "Return the arc sine of x as an angle.")
atan = decor.inv_trig(atan, "Return the arc tangent of x as an angle.")
atan2 = decor.inv_trig_yx(
    atan2,
    """Return the arc tangent of y/x as an angle.

    Unlike atan(y/x), the signs of both x and y are considered.""")
cos = decor.trig(cos, "Return the cosine of theta (an angle).")
sin = decor.trig(sin, "Return the sine of theta (an angle).")
tan = decor.trig(tan, "Return the tangent of theta (an angle).")
hypot = decor.homogeneous_copy_props(hypot)

# Angular conversion
# ------------------
del degrees, radians  # These aren't needed since angle is a quantity.

# Hyperbolic functions
# --------------------
acosh = decor.change_doc(acosh, "Return the inverse hyperbolic cosine of x.")
asinh = decor.change_doc(asinh, "Return the inverse hyperbolic sine of x.")
atanh = decor.change_doc(atanh, "Return the inverse hyperbolic tangent of x.")
# cosh, sinh, and tanh are ok as they are.

# Special functions
# -----------------
# erf, erfc, gamma, and lgamma are ok as they are.

del absolute_import, Fraction, decor




from unittest import TestCase

from scimath.physical_quantities.util import dict_add, dict_sub, dict_mul, \
    dict_div, format_expansion, unicode_powers, tex_powers, name_powers


class DictArithmeticTest(TestCase):
    def test_add(self):
        a = {'a': 3.0, 'b': -4.0, 'd': 2.0}
        b = {'a': 1.5, 'c': 12.0, 'd': -2.0}
        assert dict_add(a, b) == {'a': 4.5, 'b': -4.0, 'c': 12.0}

    def test_sub(self):
        a = {'a': 3.0, 'b': -4.0, 'd': 2.0}
        b = {'a': 1.5, 'c': 12.0, 'd': 2.0}
        assert dict_sub(a, b) == {'a': 1.5, 'b': -4.0, 'c': -12.0}

    def test_mul(self):
        a = {'a': 2.0, 'b': -4.0}
        n = 1.5
        assert dict_mul(a, n) == {'a': 3.0, 'b': -6.0}

    def test_zero_mul(self):
        a = {'a': 2.0, 'b': -4.0}
        n = 0.0
        assert dict_mul(a, n) == {}

    def test_div(self):
        a = {'a': 2.0, 'b': -4.0}
        n = 0.5
        assert dict_div(a, n) == {'a': 4.0, 'b': -8.0}


class FormatTest(TestCase):
    def setUp(self):
        self.coulomb = {"A": 1.0, 's': 1.0}
        self.joule = {"kg": 1.0, "m": 2.0, "s": -2.0}
        self.volt = {"kg": 1.0, "m": 2.0, "s": -3.0, "A": -1.0}
        self.hertz = {"s": -1.0}
        self.ohm = {u"\u2126": 1.0}
        self.siemens = {u"\u2126": -1.0}

    def test_python_defaults(self):
        assert format_expansion(self.volt) == "A**-1*kg*m**2*s**-3"

    def test_python_division(self):
        assert format_expansion(self.volt, div=True) == "kg*m**2/(A*s**3)"

    def test_python_division_simple_denominator(self):
        assert format_expansion(self.joule, div=True) == "kg*m**2/s**2"

    def test_python_division_no_denominator(self):
        assert format_expansion(self.coulomb, div=True) == "A*s"

    def test_python_division_no_numerator_simple_denominator(self):
        assert format_expansion(self.hertz, div=True) == "1/s"

    def test_python_division_no_numerator_no_denominator(self):
        assert format_expansion({}, div=True) == "1"

    def test_python_rational_power(self):
        assert format_expansion({'m': 0.5}) == "m**0.5"

    def test_unicode(self):
        assert format_expansion(self.volt, mul=" ",
                                pow_func=unicode_powers) == u"A\u207B\u00B9 kg m\u00B2 s\u207B\u00B3"

    def test_unicode_symbol(self):
        assert format_expansion(self.volt, mul=" ",
                                pow_func=unicode_powers) == u"A\u207B\u00B9 kg m\u00B2 s\u207B\u00B3"

    def test_unicode_division(self):
        assert format_expansion(self.volt, mul=" ", div=True,
                                pow_func=unicode_powers) == u"kg m\u00B2/(A s\u00B3)"

    def test_unicode_division_simple_denominator(self):
        assert format_expansion(self.joule, mul=" ", div=True,
                                pow_func=unicode_powers) == u"kg m\u00B2/s\u00B2"

    def test_unicode_division_no_denominator(self):
        assert format_expansion(self.coulomb, mul=" ", div=True,
                                pow_func=unicode_powers) == u"A s"

    def test_unicode_symbol(self):
        assert format_expansion(self.siemens, mul=" ",
                                pow_func=unicode_powers) == u"\u2126\u207B\u00B9"

    def test_unicode_symbol_division(self):
        assert format_expansion(self.siemens, mul=" ", div=True,
                                pow_func=unicode_powers) == u"1/\u2126"

    def test_unicode_rational_power(self):
        assert format_expansion({'m': 0.5}, mul=" ",
                                pow_func=unicode_powers) == u"m^0.5"

    def test_TeX_defaults(self):
        assert format_expansion(self.volt, mul="\/",
                                pow_func=tex_powers) == "A^{-1}\/kg\/m^{2}\/s^{-3}"

    def test_TeX_division(self):
        assert format_expansion(self.volt, mul="\/", div=True,
                                pow_func=tex_powers) == "kg\/m^{2}/(A\/s^{3})"

    def test_TeX_division_simple_denominator(self):
        assert format_expansion(self.joule, mul="\/", div=True,
                                pow_func=tex_powers) == "kg\/m^{2}/s^{2}"

    def test_TeX_division_no_denominator(self):
        assert format_expansion(self.coulomb, mul="\/", div=True,
                                pow_func=tex_powers) == "A\/s"

    def test_TeX_division_no_numerator_simple_denominator(self):
        assert format_expansion(self.hertz, mul="\/", div=True,
                                pow_func=tex_powers) == "1/s"

    def test_unicode_rational_power(self):
        assert format_expansion({'m': 0.5}, mul="\/",
                                pow_func=tex_powers) == "m^{0.5}"

    def test_name_defaults(self):
        assert format_expansion(self.volt, mul=" ", empty_numerator="1",
                                div_symbol=" per ", group_symbols=("", ""),
                                pow_func=name_powers) == "A to the -1 kg square m s to the -3"

    def test_name_division(self):
        assert format_expansion(self.volt, mul=" ", empty_numerator="1",
                                div_symbol=" per ", group_symbols=("", ""),
                                div=True,
                                pow_func=name_powers) == "kg square m per A cubic s"

    def test_name_division_no_numerator_simple_denominator(self):
        assert format_expansion(self.hertz, mul=" ", empty_numerator="",
                                div_symbol=" per ", group_symbols=("", ""),
                                div=True,
                                pow_func=name_powers) == " per s"



#!/usr/bin/env python
"""Clean, build, and release the HTML documentation for natu.
"""

# Standard pylint settings for this project:
# pylint: disable=I0011, C0302, C0325, R0903, R0904, R0912, R0913, R0914, R0915,
# pylint: disable=I0011, W0141, W0142

# Other:
# pylint: disable=I0011, C0103, E0611, F0401, W0621

import os
import shutil
import sys

from sh import git, python, sphinx_build, ErrorReturnCode_1, ErrorReturnCode_128
from glob import glob
from collections import namedtuple

from natu import util
from natu import groups

BUILD_DIR = 'build/html'

# Template for automatic Sphinx documentation of a module
TEMPLATE = """:orphan:

:mod:`{package}`
{underline}

.. automodule:: {package}
   :members:
   :undoc-members:
   :show-inheritance:"""

def update_groups():
    """Update the Sphinx ReST files for the submodules of natu.groups.
    """

    # Remove the existing files.
    files = glob('natu.groups.*.rst')
    for f in files:
        os.remove(f)

    # Create new files.
    packages = util.list_packages(groups)
    packages.remove('natu.groups')

    for package in packages:
        with open("%s.rst" % package, 'w') as f:
            f.write(TEMPLATE.format(package=package,
                                    underline='='*(len(package) + 7)))

def html():
    """Build/make the HTML documentation.
    """

    # Update the download link.
    try:
        commit = git('rev-list', '--tags', '--max-count=1').stdout.rstrip()
        lastversion = git.describe('--tags', commit).stdout.rstrip()
        # This is simpler but doesn't always return the latest tag:
        # lastversion = git.describe('--tag', abbrev=0).stdout.rstrip()
    except ErrorReturnCode_128:
        pass  # No tags recorded; leave download link as is
    else:
        date = git.log('-1', lastversion,
                       date='short', format='%ad').stdout[8:18]
        rpls = [(r'(natu)-.*(\.tar)', r'\1-%s\2' % lastversion[1:]),
                (r'(Latest version<br>\().*(\)</a>)',
                 r'\1%s, %s\2' % (lastversion, date)),
               ]
        util.replace('_templates/download.html', rpls)

    # Update the Sphinx ReST files for the submodules of natu.groups.
    update_groups()

    # Build the documentation.
    make_dirs()
    sphinx = sphinx_build.bake(b='html', d='build/doctrees')
    print(sphinx('.', BUILD_DIR))

    # Spellcheck.
    if util.yes("Do you want to spellcheck the HTML documentation (y/n)?"):
        spellcheck()


def clean():
    """Clean/remove the built documentation.
    """
    shutil.rmtree('build', ignore_errors=True)


def make_dirs():
    """Create the directories required to build the documentation.
    """

    # Create regular directories.
    build_dirs = ['build/doctrees', 'build/html']
    for d in build_dirs:
        try:
            os.makedirs(d)
        except OSError:
            pass

def release():
    """Release/publish the documentation to the webpage.
    """
    # Save the current state.
    branch = git('rev-parse', '--abbrev-ref', 'HEAD').stdout.rstrip()
    git.stash('save', "Work in progress while updating gh-pages branch")

    # Check out the gh-pages branch.
    try:
        git.checkout('gh-pages')
    except ErrorReturnCode_128:  # Create the branch if necessary.
        git.checkout('-b', 'gh-pages')

    # Remove the existing files in the base folder.
    extensions = ['*.html', '*.inv']
    fnames = util.multiglob('..', extensions)
    for fname in fnames:
        os.remove(fname)

    # Copy the new files to the base folder.
    fnames = util.multiglob(BUILD_DIR, extensions)
    for fname in fnames:
        shutil.copy(fname, '..')

    # Track the new files.
    fnames = util.multiglob('..', extensions)
    git.add(*fnames)

    # Copy but rename the folders referenced in the HTML files.
    # Github only recognizes images, stylesheets, and javascripts as folders.
    folders = [('_images', 'images'),
               ('_static', 'javascripts'),
              ]
    for (src, dst) in folders:
        dst = os.path.join('..', dst)
        # Remove the existing folder.
        shutil.rmtree(dst, ignore_errors=True)
        # Copy the new folder.
        shutil.copytree(os.path.join(BUILD_DIR, src), dst)
        # Track the new folder.
        git.add(dst)
    # Update the HTML files to reference the new folder names.
    html_fnames = glob(os.path.join('..', '*.html'))
    util.replace(html_fnames, folders)

    # Update the sitemap.
    print(python('sitemap_gen.py', config="sitemap_conf.xml"))

    # Commit the changes.
    try:
        git.commit('-a', m="Rebuilt documentation")
    except ErrorReturnCode_1:
        pass  # No changes to commit

    # If desired, rebase and push the changes to origin.
    print("The gh-pages branch has been updated and is currently checked out.")
    if util.yes("Do you want to rebase it and push the changes to "
                "origin (y/n)?"):
        git.rebase('-i', 'origin/gh-pages')
        git.push.origin('gh-pages')

    # Return to the original state.
    git.checkout(branch)
    try:
        git.stash.pop()
    except ErrorReturnCode_1:
        pass  # No stash was necessary in the first place.
    print("Now back on " + branch)


def spellcheck():
    """Spellcheck the HTML docs.
    """
    # Options
    wordfile = os.path.abspath('natu.pws') # Name of custom word file
    html_files = glob('build/html/*.html') # Names of the HTML files

    print("If there are misspellings, fix them in the Python or ReST "
          "source---not just in the HTML files.")

    # Remove unused words from the custom word file.
    def read(fname):
        """Read all text from a file."""
        with open(fname, 'r') as f:
            return f.read()
    html = "".join(map(read, html_files))
    with open(wordfile, 'r') as f:
        head = f.readline()
        lines = f.readlines()
    with open(wordfile, 'w') as f:
        f.write(head)
        for line in lines:
            if html.find(line.rstrip('\n').rstrip('\r')) != -1:
                f.write(line)

    # Check the spelling.
    for page in html_files:
        if os.system('aspell --dont-backup --personal={0} '
                     '-c {1}'.format(wordfile, page)):
            raise SystemError("aspell (http://aspell.net/) must be installed.")


Action = namedtuple("Action", ['f', 'description'])
ACTIONS = {'clean'   : Action(clean, "Clean/remove the built documentation."),
           'build'   : Action(html, "Build/make the HTML documentation."),
           'html'    : Action(html, "Build/make the HTML documentation."),
           'release' : Action(release, ("Release/publish the documentation to "
                                        "the webpage.")),
          }


def funcs_str():
    """Return a string listing the valid functions and their descriptions.
    """
    return "\n".join(["  %s: %s" % (action.ljust(10), function.description)
                      for (action, function) in ACTIONS.items()])


# Main
if len(sys.argv) != 2:
    raise SystemExit("Exactly one argument is required; valid args are\n"
                     + funcs_str())
ACTION = sys.argv[1]
try:
    ACTIONS[ACTION].f()
except KeyError:
    raise SystemExit("Do not know how to handle %s; valid args are\n%s"
                     % (ACTION, funcs_str()))



""" Test unit conversion functions used on input and output of functions.

    fixme: We need significant work on scalars.
"""

# Standard Library imports
import unittest

# Numeric library imports
from numpy import array, all, allclose, ndarray

# Enthought library imports
from scimath.units.unit import InvalidConversion
from scimath.units.length import feet, meters
from scimath.units.time import second

# Numerical modeling library imports
from scimath.units.api import UnitArray, UnitScalar
from scimath.units.unit_manipulation import \
    convert_units, set_units, have_some_units, strip_units

class ConvertUnitsTestCase(unittest.TestCase):
    """ ConvertUnits should pretty much leave anything without units alone
        and pass them through silently.  UnitArrays do get converted,
        and so should scalars with units (although we haven't really dealt with
        those).
    """

    ############################################################################
    # ConvertUnitsTestCase interface.
    ############################################################################

    def test_single_float(self):
        """ Does it pass a single value through correctly?
        """
        units = [None]
        result = convert_units(units, 1.0)
        self.assertEqual(1.0, result)

    def test_two_float(self):
        """ Does it pass a two values through correctly?
        """
        units = [None, None]
        result = convert_units(units, 1.0, 2.0)
        self.assertEqual([1.0, 2.0], result)

    def test_mismatch_raises_error(self):
        """ Is an exception raised if there aren't enough units specified?
        """
        self.assertRaises(ValueError, convert_units, ([None], 1.0, 2.0))

    def test_one_array(self):
        """ Does it pass a an array through correctly?
        """
        units = [None]
        a = array((1,2,3))
        result = convert_units(units, a)
        self.assertTrue(all(a==result))

    def test_two_arrays(self):
        """ Does it pass a two arrays through correctly?
        """
        units = [None, None]
        a = array((1,2,3))
        b = array((3,4,5))
        aa,bb = convert_units(units, a, b)
        self.assertTrue(all(a==aa))
        self.assertTrue(all(b==bb))

    def test_convert_array_with_units(self):
        """ Does it add units to an array correctly?

            fixme: This may be exactly what we don't want to happen!
        """
        units = [feet]
        a = array((1,2,3))
        aa = convert_units(units, a)
        self.assertTrue(all(a==aa))
        self.assertTrue(type(aa) is ndarray)

    def test_convert_unit_array(self):
        """ Does it convert an array correctly?
        """
        units = [feet]
        a = UnitArray((1,2,3),units=meters)
        aa = convert_units(units, a)
        self.assertTrue(allclose(a,aa.as_units(meters)))
        # fixme: This actually may be something we don't want.  For speed,
        #        if this were just a standard array, we would be better off.
        self.assertEqual(aa.units, feet)

    def test_convert_unit_scalar(self):
        """ Does it convert a scalar correctly?
        """
        units = [feet]
        a = UnitScalar(3.,units=meters)
        aa = convert_units(units, a)
        self.assertTrue(allclose(a,aa.as_units(meters)))
        self.assertEqual(aa.units, feet)

    def test_incompatible_array_units_raise_exception(self):
        """ Does a units mismatch raise an exception?

            fixme: Do we want this configurable?
        """
        units = [second]
        a = UnitArray((1,2,3),units=meters)
        self.assertRaises(InvalidConversion, convert_units, units, a)

    def test_incompatible_scalar_units_raise_exception(self):
        """ Does a units mismatch raise an exception?

            fixme: Do we want this configurable?
        """
        units = [second]
        a = UnitScalar(3.,units=meters)
        self.assertRaises(InvalidConversion, convert_units, units, a)

    def test_dont_convert_unit_array(self):
        """ Does it return the same object if units are the same?

            Note: This isn't required for accuracy, but it is a good
                  optimization.
        """
        units = [feet]
        a = UnitArray((1,2,3),units=feet)
        aa = convert_units(units, a)
        self.assertTrue(id(a),id(aa))

    def test_dont_convert_unit_scalar(self):
        """ Does it return the same object if units are the same?

            Note: This isn't required for accuracy, but it is a good
                  optimization.
        """
        units = [feet]
        a = UnitScalar(3.,units=feet)
        aa = convert_units(units, a)
        self.assertTrue(id(a),id(aa))

    def test_convert_different_args(self):
        """ Does it handle multiple different args correctly?
        """
        units = [feet, meters, None, feet]
        a = UnitArray((1,2,3),units=meters)
        b = array((2,3,4))
        c = 1
        d = UnitScalar(3.,units=meters)
        aa, bb, cc, dd = convert_units(units, a, b, c, d)
        self.assertTrue(allclose(a,aa.as_units(meters)))
        self.assertTrue(allclose(b,bb))
        self.assertEqual(c,cc)
        self.assertTrue(allclose(d,dd.as_units(meters)))

class SetUnitsTestCase(unittest.TestCase):


    ############################################################################
    # TestCase interface.
    ############################################################################

    def setUp(self):
        unittest.TestCase.setUp(self)

    def tearDown(self):
        unittest.TestCase.tearDown(self)

    ############################################################################
    # SetUnitsTestCase interface.
    ############################################################################

    def test_single_float(self):
        """ Does it pass a single value through correctly?
        """
        units = [None]
        result = set_units(units, 1.0)
        self.assertEqual(1.0, result)

    def test_mismatch_raises_error(self):
        """ Is an exception raised if there aren't enough units specified?
        """
        self.assertRaises(ValueError, convert_units, [None], 1.0, 2.0)

    def test_one_array(self):
        """ Does it pass a an array through correctly?
        """
        units = [None]
        a = array((1,2,3))
        result = set_units(units, a)
        self.assertTrue(all(a==result))

    def test_set_scalar_with_units(self):
        """ Does it add units to a scalar correctly?
        """
        units = [feet]
        x = 3.0
        xx = set_units(units, x)
        self.assertEqual(float(xx), x)
        self.assertEqual(xx.units, feet)

    def test_set_array_with_units(self):
        """ Does it add units to an array correctly?

            fixme: This may be exactly what we don't want to happen!
        """
        units = [feet]
        a = array((1,2,3))
        aa = set_units(units, a)
        self.assertTrue(all(a==aa))
        self.assertEqual(aa.units, feet)

    def test_set_zero_dim_array_with_units(self):
        """ Does it add units to an array with shape () correctly?

            fixme: This may be exactly what we don't want to happen!
        """
        units = [feet]
        a = array(2)
        aa = set_units(units, a)
        self.assertTrue(all(a==aa))
        self.assertEqual(aa.units, feet)
        assert isinstance(aa, UnitScalar)

    def test_set_unit_overwrite_unit_scalar(self):
        """ Does it overwrite units on a UnitScalar correctly?
        """
        units = [feet]
        x = UnitScalar(3., units=meters)
        xx = set_units(units, x)
        # FIXME:
        #     Behaves very stangely (on my machine), somethimes it fails,
        #     other times it works almost like a random generator.
        #
        # We found that set_units(units, x) has a sideffect on x which it
        # should not have.
        #
        #self.assertEqual(x, xx)
        #print x, x.units
        self.assertEqual(xx.units, feet)

    def test_set_unit_overwrite_unit_array(self):
        """ Does it overwrite units on a UnitArray correctly?
        """
        units = [feet]
        a = UnitArray((1,2,3),units=meters)
        aa = set_units(units, a)
        self.assertTrue(all(a==aa))
        self.assertEqual(aa.units, feet)
#
#    def test_raises_exception(self):
#        """ Does it return the same object if units are the same?
#
#            Note: This isn't required for accuracy, but it is a good
#                  optimization.
#        """
#        units = [feet]
#        a = UnitArray((1,2,3),units=feet)
#        aa = convert_units(units, a)
#        self.assertTrue(id(a),id(aa))
#
#    def test_convert_different_args(self):
#        """ Does it handle multiple different args correctly?
#        """
#        units = [feet, meters, None]
#        a = UnitArray((1,2,3),units=meters)
#        b = array((2,3,4))
#        c = 1
#        aa, bb, cc = convert_units(units, a, b, c)
#        self.assertTrue(allclose(a,aa.as_units(meters)))
#        self.assertTrue(allclose(b,bb))
#        self.assertEqual(c,cc)


class HaveSomeUnitsTestCase(unittest.TestCase):
    """ have_some_units should check its arguments for any
    UnitArrays/UnitScalars.
    """

    ############################################################################
    # TestCase interface.
    ############################################################################

    def setUp(self):
        # Make some useful data.
        self.unit_array = UnitArray((1,2,3), units=meters)
        self.unit_scalar = UnitScalar(1, units=meters)
        self.plain_array = array([1, 2, 3])
        self.plain_scalar = 1
        unittest.TestCase.setUp(self)

    def test_finds_one(self):
        self.assertTrue(have_some_units(self.unit_array))
        self.assertTrue(have_some_units(self.unit_scalar))

    def test_finds_multiple(self):
        self.assertTrue(have_some_units(self.unit_array, self.unit_array))
        self.assertTrue(have_some_units(self.unit_scalar, self.unit_scalar))

    def test_finds_mixed_scalar_array(self):
        self.assertTrue(have_some_units(self.unit_array, self.unit_scalar))

    def test_does_not_find_plain(self):
        self.assertFalse(have_some_units(self.plain_array))
        self.assertFalse(have_some_units(self.plain_scalar))

    def test_does_not_find_mixed_plain(self):
        self.assertFalse(have_some_units(self.plain_array, self.plain_scalar))

    def test_finds_any_unitted(self):
        self.assertTrue(have_some_units(self.unit_array, self.plain_array, self.plain_scalar))
        self.assertTrue(have_some_units(self.plain_array, self.unit_array, self.plain_scalar))
        self.assertTrue(have_some_units(self.unit_scalar, self.plain_array, self.plain_scalar))
        self.assertTrue(have_some_units(self.plain_array, self.unit_scalar, self.plain_scalar))


class StripUnitsTestCase(unittest.TestCase):
    """ strip_units should remove units from UnitArrays/UnitScalars.
    """

    def setUp(self):
        # Make some useful data.
        self.unit_array = UnitArray((1,2,3), units=meters)
        self.unit_scalar = UnitScalar(1, units=meters)
        self.plain_array = array([1, 2, 3])
        self.plain_scalar = 1
        unittest.TestCase.setUp(self)

    def test_strip_units_one_arg(self):
        self.assertFalse(isinstance(strip_units(self.unit_array),
            (UnitArray, UnitScalar)))
        self.assertFalse(isinstance(strip_units(self.unit_scalar),
            (UnitArray, UnitScalar)))
        self.assertFalse(isinstance(strip_units(self.plain_array),
            (UnitArray, UnitScalar)))
        self.assertFalse(isinstance(strip_units(self.plain_scalar),
            (UnitArray, UnitScalar)))

        # Check for stupidity when returning only one argument.
        self.assertFalse(isinstance(strip_units(self.unit_scalar), tuple))
        self.assertFalse(isinstance(strip_units(self.plain_scalar), tuple))

    def test_strip_units_multi_arg(self):
        outs = strip_units(self.unit_array, self.unit_scalar)
        self.assertEquals(len(outs), 2)
        for x in outs:
            self.assertFalse(isinstance(x, (UnitArray, UnitScalar)))

        outs = strip_units(self.plain_array, self.plain_scalar)
        self.assertEquals(len(outs), 2)
        for x in outs:
            self.assertFalse(isinstance(x, (UnitArray, UnitScalar)))

        outs = strip_units(self.unit_array, self.unit_scalar, self.plain_array,
            self.plain_scalar)
        self.assertEquals(len(outs), 4)
        for x in outs:
            self.assertFalse(isinstance(x, (UnitArray, UnitScalar)))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
"""Contains *PREFIXES*, a dictionary of SI prefixes
"""
# [BIPM2006]: International Bureau of Weights and Measures (BIPM), "The
#             International System of Units (SI)," 8th ed., 2006.

# [BIPM2006, Table 5]
PREFIXES = dict(Y=1e24,   # yotta
                Z=1e21,   # zetta
                E=1e18,   # exa
                P=1e15,   # peta
                T=1e12,   # tera
                G=1e9,    # giga
                M=1e6,    # mega
                k=1e3,    # kilo
                h=100,    # hecto
                da=10,    # deca
                d=0.1,    # deci
                c=0.01,   # centi
                m=1e-3,   # milli
                u=1e-6,   # micro
                n=1e-9,   # nano
                p=1e-12,  # pico
                f=1e-15,  # femto
                a=1e-18,  # atto
                z=1e-21,  # zepto
                y=1e-24)  # yocto



#!/usr/bin/python
"""Aliases for the `SI units`_


.. _SI units: http://en.wikipedia.org/wiki/International_System_of_Units
"""
# pylint: disable=I0011, C0103, E0611

from . import _update_module
from ..units import _units

# [BIPM2006, Table 1]: SI base units
# g is included so that it can be used with other prefixes than k.
units = {symbol: _units[symbol] for symbol in 'm kg g s A K mol cd'.split()}

# [BIPM2006, Table 3]: Coherent derived units in SI with special names and
# symbols
units.update({symbol: _units[symbol] for symbol in
              ('rad sr Hz N Pa J W C V F ohm S Wb T H degC lm lx Bq Gy '
               'Sv kat'.split())})

_update_module(__name__, units)



#!/usr/bin/python
"""Natural units in Python
"""
__author__ = "Kevin Davies"
__email__ = "kdavies4@gmail.com"
__copyright__ = ("Copyright 2013-2014, Kevin Davies, Hawaii Natural Energy "
                 "Institute, and Georgia Tech Research Corporation")
__license__ = "BSD-compatible (see LICENSE.txt)"
__version__ = '0.1.0-a5'



# Standard Library imports
import unittest
import time

# Numeric library imports
import numpy
from numpy import arange, allclose, array, all #@UnresolvedImport
from numpy.testing import assert_array_almost_equal

# Enthought library imports
from traits.testing.api import doctest_for_module, skip
from scimath.units.length import feet, meters
from scimath.units.time import second
from scimath.units.unit_parser import unit_parser

# Numerical modeling library imports
import scimath.units.has_units as has_units_
from scimath.units.api import has_units, UnitArray, UnitScalar

class HasUnitsDocTestCase(doctest_for_module(has_units_)):
    pass

class HasUnitsTestCase(unittest.TestCase):

    ############################################################################
    # TestCase interface.
    ############################################################################

    def setUp(self):
        unittest.TestCase.setUp(self)

    def tearDown(self):
        unittest.TestCase.tearDown(self)

    def failUnlessEqual(self, first, second, msg=None):
        """Fail if the two objects are unequal as determined by the '=='
           operator.

           We've overloaded this here to handle arrays.
        """
        if not all(first == second):
            raise self.failureException, \
                  (msg or '%r != %r' % (first, second))

    assertEqual = failUnlessEqual

    ############################################################################
    # HasUnitsTestCase interface.
    ############################################################################

    def test_wrapped_behaves_like_unwrapped(self):
        """ Does wrapped function behave like non-wrapped function.

        """
        def func(value):
            return value

        decorater = has_units()
        func_wrapped = decorater(func)

        self.assertEqual(func(1),func_wrapped(1))

    def test_wrapped_with_input_units(self):
        """ Does wrapped function with inputs behave like non-wrapped function?

        """

        def func(value):
            return value

        decorater = has_units(inputs="value: a value: units=m/s;")
        func_wrapped = decorater(func)

        self.assertEqual(func(1),func_wrapped(1))

    def test_wrapped_with_output_units(self):
        """ Does wrapped function with outputs behave like non-wrapped function?

        """

        def func(value):
            return value

        decorater = has_units(outputs="value: a value: units=m/s;")
        func_wrapped = decorater(func)

        arg = array((1,2,3))
        std = func(arg)
        actual = func_wrapped(arg)
        self.assertEqual(std,actual)

    def test_wrapped_with_two_output_units(self):
        """ Are two outputs handled correctly?

        """

        def func(v1, v2):
            return v1, v2

        decorater = has_units(outputs="""v1: a value: units=m/s;
                                        v2: another value: units=ft/s;
                                     """)
        func_wrapped = decorater(func)

        arg = array((1,2,3))
        std_v1, std_v2 = func(arg, arg+1)
        actual_v1, actual_v2 = func_wrapped(arg, arg+1)
        self.assertEqual(std_v1, actual_v1)
        self.assertEqual(std_v2, actual_v2)

    def test_wrapped_with_units_decorated(self):

        def func(value):
            return value

        @has_units(inputs="value: a value: units=m/s;")
        def func_wrapped(value):
            return value


        self.assertEqual(func(1),func_wrapped(1))

    def test_unit_array_with_units_decorated(self):

        def func(value):
            return value

        @has_units(inputs="value: a value: units=m/s;")
        def func_wrapped(value):
            return value

        a = UnitArray(arange(100), units=meters/second)
        self.assertTrue(allclose(func(a).as_units(meters/second),
                                 func_wrapped(a)))

        a = UnitArray(arange(100), units=feet/second)
        self.assertTrue(allclose(func(a).as_units(meters/second),
                                 func_wrapped(a)))

    def test_unit_array_with_decorated_docstring_function(self):
        """Does has_units wrap with docstring work ?
        """

        def addfunc(a,b):
            return a+b

        @has_units
        def add(a,b):
            ''' Add two arrays in ft/s and convert them to m/s.

            Parameters
            ----------
            a : array : units=ft/s
                An array
            b : array : units=ft/s
                Another array

            Returns
            -------
            c : array : units=m/s
                c = a + b
            '''
            return (a+b)*0.3048

        a = UnitArray(arange(100), units=feet/second)
        b = UnitArray(arange(100)**2, units=feet/second)
        self.assertTrue(allclose(addfunc(a,b).as_units(meters/second),
                                 add(a,b)))

        a = UnitArray(arange(100), units=meters/second)
        b = UnitArray(arange(100)**2, units=meters/second)
        self.assertTrue(allclose(addfunc(a,b).as_units(meters/second),
                                 add(a,b)))

    def test_unit_array_with_decorated_docstring_and_inputted_parameters(self):
        """Does has_units wrap with expanded docstring and inputting
        parameters work the same ?
        """

        @has_units(inputs="a:an array:units=ft/s;b:array:units=ft/s",
                   outputs="c:an array:units=m/s")
        def add(a,b):
            " Add two arrays in ft/s and convert them to m/s. "
            return (a+b)*0.3048

        @has_units
        def add_doc(a, b):
            ''' Add two arrays in ft/s and convert them to m/s.

            Parameters
            ----------
            a : array : units=ft/s
                An array
            b : array : units=ft/s
                Another array

            Returns
            -------
            c : array : units=m/s
                c = a + b
            '''
            return (a+b)*0.3048

        a = UnitArray(arange(100), units=feet/second)
        b = UnitArray(arange(100)**2, units=feet/second)
        self.assertTrue(allclose(add_doc(a,b),
                                 add(a,b)))

        a = UnitArray(arange(100), units=meters/second)
        b = UnitArray(arange(100)**2, units=meters/second)
        self.assertTrue(allclose(add_doc(a,b),
                                 add(a,b)))

    def test_wrapped_adds_summary(self):
        """ Is summary information added correctly?
        """
        summary = "a function"
        @has_units(summary=summary)
        def func(value):
            return value

        self.assertTrue(func.summary==summary)

    def test_wrapped_adds_doc(self):
        """ Is doc information added correctly?
        """
        doc = "documenation about the function"
        @has_units(doc=doc)
        def func(value):
            return value

        self.assertTrue(func.doc==doc)

    def test_wrapped_adds_inputs(self):
        """ Are input specifications added correctly?
        """

        @has_units(inputs="v1:a value:units=m/s")
        def func(v1, v2):
            return v1+v2

        self.assertTrue(len(func.inputs)==2)

        # Does the 1st variable have its name and units assigned correctly?
        self.assertTrue(func.inputs[0].name=='v1')
        self.assertTrue(func.inputs[0].units==unit_parser.parse_unit('m/s'))

        # Was the 2nd (unspecified) input given an input variable?
        self.assertTrue(func.inputs[1].name=='v2')
        self.assertTrue(func.inputs[1].units==None)

    def test_wrapped_adds_outputs(self):
        """ Are output specifications added correctly?
        """

        @has_units()
        def func(v1, v2):
            return v1+v2

        self.assertTrue(len(func.outputs)==1)

        # Does the output have its name and units assigned correctly?
        self.assertTrue(func.outputs[0].name=='result')
        self.assertTrue(func.outputs[0].units==None)

    def test_wrapped_adds_outputs2(self):
        """ Are output specifications added correctly?
        """

        @has_units(outputs="out: out desc:units=m/s")
        def func(v1, v2):
            return v1+v2

        self.assertTrue(len(func.outputs)==1)

        # Does the output have its name and units assigned correctly?
        self.assertTrue(func.outputs[0].name=='out')
        self.assertTrue(func.outputs[0].units==unit_parser.parse_unit('m/s'))

    def _time_conversion_decorated(self, inputs, array_size=1000,iters=1000):
        """ Wrapped vs. normal have slowdown<1.2 (no conversion)?
        """
        def func(a,b,c,d):
            return a+a,b+b,c+c,d+d

        @has_units(inputs=inputs)
        def func_wrapped(a,b,c,d):
            return a+a,b+b,c+c,d+d

        a = UnitArray(arange(array_size), units=meters/second)
        #aa = arange(array_size)

        N = iters
        t1 = time.clock()
        for i in range(N): #@UnusedVariable
            w,x,y,z = func(a,a,a,a) #@UnusedVariable
        t2 = time.clock()
        standard = t2 - t1

        t1 = time.clock()
        for i in range(N): #@UnusedVariable
            w,x,y,z = func_wrapped(a,a,a,a) #@UnusedVariable
        t2 = time.clock()
        wrapped = t2-t1

        b=a
        c=a
        d=a
        t1 = time.clock()
        for i in range(N): #@UnusedVariable
            w,x,y,z = a+a,b+b,c+c,d+d #@UnusedVariable
        t2 = time.clock()
        bare = t2-t1

        slowdown = wrapped/standard
        msg = ("call/s = %s, slowdown = %s; call/s = %s, slowdown = %s" %
                  (N/wrapped, slowdown, N/bare, wrapped/bare))
        self.assertTrue(slowdown<1.2, msg)

    @skip
    def test_time_no_conversion_decorated(self):
        """ Wrapped vs. normal have slowdown<1.2 (no conversion)?
        """
        inputs="""a: description of a: units=m/s;
                  b: description of b: units=m/s;
                  c: description of c: units=m/s;
                  d: description of d: units=m/s;
               """
        self._time_conversion_decorated(inputs=inputs)

    @skip
    def test_time_conversion_decorated(self):
        """ Wrapped vs. normal have slowdown<1.2 (with conversion)?
        """
        inputs="""a: description of a: units=ft/s;
                  b: description of b: units=ft/s;
                  c: description of c: units=ft/s;
                  d: description of d: units=ft/s;
               """
        self._time_conversion_decorated(inputs=inputs)

# Some functions to play with.
def foo(x, y):
    """ Foo

    Parameters
    ----------
    x : scalar : units=m
        X
    y : scalar : units=s
        Y

    Returns
    -------
    z : scalar : units=m/s
    """
    assert not isinstance(x, (UnitArray, UnitScalar))
    assert not isinstance(y, (UnitArray, UnitScalar))
    z = x / y
    return z

foo_with_units = has_units(foo)

def bar(x, y):
    """ Bar

    Parameters
    ----------
    x : scalar : units=m
        X
    y : scalar : units=s
        Y

    Returns
    -------
    z : scalar : units=m
    """

    if y > 1:
        z = x - x
    else:
        z = x + x
    return z


vec_bar_with_units = has_units(numpy.vectorize(bar))


class HasUnitsDecoratorTestCase(unittest.TestCase):

    def setUp(self):
        # Make some data to play with.
        self.meter_array = UnitArray([1.,2,3], units=meters)
        self.second_array = UnitArray([3.,2,1], units=second)
        self.feet_array = UnitArray([4.,5,6], units=feet)
        self.meter_scalar = UnitScalar(1., units=meters)
        self.second_scalar = UnitScalar(3., units=second)
        self.feet_scalar = UnitScalar(4., units=feet)
        unittest.TestCase.setUp(self)

    def test_decorator_plays_nice(self):
        self.assertEquals(foo_with_units.__module__, foo.__module__)
        self.assertEquals(foo_with_units.__doc__, foo.__doc__)
        self.assertEquals(foo_with_units.__name__, foo.__name__)

    def test_input_variables_parsed(self):
        inputs = foo_with_units.inputs
        self.assertEquals(len(inputs), 2)
        self.assertEquals(inputs[0].name, 'x')
        self.assertEquals(inputs[0].units, meters)
        self.assertEquals(inputs[1].name, 'y')
        self.assertEquals(inputs[1].units, second)

    def test_output_variables_parsed(self):
        outputs = foo_with_units.outputs
        self.assertEquals(len(outputs), 1)
        self.assertEquals(outputs[0].name, 'z')
        self.assertEquals(outputs[0].units, meters/second)

    def test_no_internal_units_array(self):
        z = foo_with_units( self.meter_array, self.second_array)
        self.assertTrue(isinstance(z, UnitArray))
        self.assertEquals(z.units, meters/second)

    def test_no_internal_units_scalar(self):
        z = foo_with_units( self.meter_scalar, self.second_scalar)
        self.assertTrue(isinstance(z, UnitScalar))
        self.assertEquals(z.units, meters/second)

    def test_feet(self):
        z = foo_with_units( self.feet_array, self.second_array)
        self.assertTrue(isinstance(z, UnitArray))
        self.assertEquals(z.units, meters/second)
        assert_array_almost_equal(z, numpy.array([ 0.4064,  0.762 ,  1.8288]))
        z = foo_with_units( self.feet_scalar, self.second_scalar)
        self.assertTrue(isinstance(z, UnitScalar))
        self.assertEquals(z.units, meters/second)
        assert_array_almost_equal(z, 0.4064)

    def test_v_decorator_plays_nice(self):
        self.assertEquals(vec_bar_with_units.__module__, bar.__module__)
        self.assertEquals(vec_bar_with_units.__doc__, bar.__doc__)
        self.assertEquals(vec_bar_with_units.__name__, bar.__name__)

    def test_v_input_variables_parsed(self):
        inputs = vec_bar_with_units.inputs
        self.assertEquals(len(inputs), 2)
        self.assertEquals(inputs[0].name, 'x')
        self.assertEquals(inputs[0].units, meters)
        self.assertEquals(inputs[1].name, 'y')
        self.assertEquals(inputs[1].units, second)

    def test_v_output_variables_parsed(self):
        outputs = vec_bar_with_units.outputs
        self.assertEquals(len(outputs), 1)
        self.assertEquals(outputs[0].name, 'z')
        self.assertEquals(outputs[0].units, meters)

    def test_v_no_internal_units_array(self):
        z = vec_bar_with_units( self.meter_array, self.second_array)
        self.assertTrue(isinstance(z, UnitArray))
        self.assertEquals(z.units, meters)

    def test_v_no_internal_units_scalar(self):
        z = vec_bar_with_units( self.meter_scalar, self.second_scalar)
        self.assertTrue(isinstance(z, UnitScalar))
        self.assertEquals(z.units, meters)

    def test_v_feet(self):
        z = vec_bar_with_units( self.feet_array, self.second_array)
        self.assertTrue(isinstance(z, UnitArray))
        self.assertEquals(z.units, meters)
        assert_array_almost_equal(z, numpy.array([ 0.0,  0.0 ,  3.6576]))
        z = vec_bar_with_units( self.feet_scalar, self.second_scalar)
        self.assertTrue(isinstance(z, UnitScalar))
        self.assertEquals(z.units, meters)
        assert_array_almost_equal(z, 0.0)

if __name__ == '__main__':
    # profile the test suite.
    import hotshot, hotshot.stats
    prof = hotshot.Profile("convert.prof")
    prof.runcall(unittest.main)
    prof.close()
    stats = hotshot.stats.load("convert.prof")
    stats.strip_dirs()
    stats.sort_stats('time', 'calls')
    stats.print_stats(20)
    #import sys
    #unittest.main(argv=sys.argv)



#!/usr/bin/python
"""Aliases for constants and units of force, with support for prefixes
"""
# pylint: disable=I0011, E0611

from . import _update_module
from ..units import _units, N

_update_module(__name__, _units, N.dimension)



#!/usr/bin/python
# -*- coding: utf-8 -*-
r"""Configuration settings

This module contains the following settings (defaults in parentheses):

- *definitions* (`base-SI.ini <base-ini.html>`_, `derived.ini
  <derived-ini.html>`_, `BIPM.ini <BIPM-ini.html>`_, and `other.ini
  <other-ini.html>`_ from the installation directory) - List of files that
  define the constants and units

- *use_quantities* (*True*) - *True* to use quantities to track dimensions and
  display units

     If *use_quantities* is *False*, then the constants and scalar units are
     :class:`float` instances rather than :class:`~natu.core.Quantity` and
     :class:`~natu.core.ScalarUnit` instances.  This reduces the computational
     overhead of the :mod:`natu` module to nearly zero while still allowing
     variables to be specified using various units.  However, this disables
     dimension checking and the string formatting of quantities as the product
     of a number and a unit, so it is probably best to leave *use_quantities*
     set to *True* until you have validated your unit-dependent code.

- *simplification_level* (2) - Number of non-minimizing substitutions that can
  be made in seeking the best display unit

     A higher number increases the likelihood that the simplest display unit
     will be found, but it also increases the time required to process
     :func:`str`, :func:`print`, :func:`format`, and related functions.

- *unit_replacements*
  (in LaTeX_ ('L' format code):
  'deg' → '^{\\circ}',
  'ohm' → '\\Omega', and
  'angstrom' → '\\AA';
  in Unicode_ ('U' format code):
  'deg' → '°',
  'ohm' → 'Ω', and
  'angstrom' → 'Å'): Dictionary of special replacements for formatting a unit
  string

     Each key is a format code and each value is a list of tuples of 1) the
     string to be replaced and 2) the string that it should be replaced with.

Import this module and change these settings as desired before importing any
other :mod:`natu` submodules.

**Examples:**

To turn off unit simplification:

>>> from natu import config
>>> config.simplification_level = 0

or to append a custom file of unit definitions:

>>> config.definitions.append('custom.ini')

Then use the rest of :mod:`natu`, for instance

>>> from natu.units import *

.. testcleanup::
   >>> print(kg/m/s)
   kg/(m*s)


.. _Unicode: https://en.wikipedia.org/wiki/Unicode
.. _LaTeX: http://www.latex-project.org/
"""
# pylint: disable=I0011, C0103

# List of unit definition files
from os import path
dname = path.dirname(__file__)
definitions = [path.join(dname, fname) for fname in
               ['base-SI.ini', 'derived.ini', 'BIPM.ini', 'other.ini']]
del path, dname

# True to track dimensions and display units:
use_quantities = True

# Number of non-minimizing substitutions that can be made in seeking the
# best display unit:
simplification_level = 2

# Dictionary of special replacements for formatting a unit string
unit_replacements = {'L':  # In LaTeX
                     [('deg', r'^{\circ}'),
                      ('ohm', r'\Omega'),
                      ('angstrom', r'\AA')],
                     'U':  # In Unicode format
                     [('deg', '°'),
                      ('ohm', 'Ω'),
                      ('angstrom', 'Å')],
                    }



#!/usr/bin/python
r"""Module with all units from the `definition files`_ listed in
:attr:`natu.config.definitions`

This module cannot be reloaded.  This prevents conflicts that could arise if the
units were redefined with different `base constants <base-ini.html>`_.


.. _definition files: definitions.html
"""
__author__ = "Kevin Davies"
__email__ = "kdavies4@gmail.com"
__copyright__ = ("Copyright 2013-2014, Kevin Davies, Hawaii Natural Energy "
                 "Institute, and Georgia Tech Research Corporation")
__license__ = "BSD-compatible (see LICENSE.txt)"

if __name__ == '__main__':
    # Test the contents of this file.
    import doctest
    doctest.testmod()
else:
    # Replace the module with a UnitsModule for dynamic unit prefixing.
    from sys import modules
    from .core import UnitsModule
    from .config import definitions
    modules[__name__] = UnitsModule(modules[__name__], definitions)



#!/usr/bin/python
"""Aliases for physical constants, with support for prefixes
"""
# pylint: disable=I0011, C0103, E0611

from . import _update_module
from ..units import _units
from ..core import Quantity, Unit

# Constants are quantities but not units.
units = {symbol: quantity for symbol, quantity in _units.items()
         if isinstance(quantity, Quantity) and not isinstance(quantity, Unit)}
try:
    units.update({'k_A': _units['k_A']}) # Ampere's constant is a ScalarUnit.
except KeyError:
    pass

_update_module(__name__, units)



# Standard library imports
from cPickle import dumps, loads
import timeit
import unittest
import operator

from nose import SkipTest

# Numeric library imports
import numpy
from numpy import all, array, sqrt

# Enthought Library imports
import scimath.units as units
from traits.testing.api import skip
from scimath.units.length import meters, feet
from scimath.units.time import second, seconds
from scimath.units.unit import InvalidConversion, dimensionless

# Numerical modeling library imports
from scimath.units.api import UnitArray, UnitScalar

class UnitArrayTestCase(unittest.TestCase):

    ############################################################################
    # unittest.TestCase interface
    ############################################################################

    def setUp(self):
        unittest.TestCase.setUp(self)

    def tearDown(self):
        unittest.TestCase.tearDown(self)


    ############################################################################
    # Test construction from arrays
    ############################################################################

    def test_create_1d_from_list(self):
        """ Create an array from a 1D list.
        """
        a = UnitArray([1,2,3])
        assert(a[0] == 1)

    def test_create_2d_from_list(self):
        """ Create an array from a 2D list.
        """
        a = UnitArray([[1,2,3],
                 [4,5,6]])
        assert(a[1,2] == 6)


    ############################################################################
    # Test mathematical operations
    ############################################################################

    def test_sin(self):
        """ Unary operation on a UnitArray should return a UnitArray.
        """
        a = UnitArray([1,2,3])
        res = numpy.sin(a) #@UndefinedVariable
        assert (type(res) is UnitArray)

    def test_add(self):
        """ Binary operation of array + UnitArray should return a UnitArray.
        """
        ary = numpy.array((1,2,3))
        unit_ary = UnitArray(ary)
        res = unit_ary + ary
        assert type(res) is UnitArray

    def test_array_equality(self):
        """ Rich comparison of array == UnitArray should compare the data
        """
        ary = numpy.array((1,2,3))
        unit_ary = UnitArray(ary)
        res = unit_ary + ary
        assert numpy.all(res == (ary+ary))


    ############################################################################
    # Tests for units
    ############################################################################

    def test_units_default(self):
        """ Are their units on this UnitArray.
        """
        ary = numpy.array((1,2,3))
        unit_ary = UnitArray(ary)
        self.assertEqual(unit_ary.units, None)

    def test_units_init(self):
        """ Are their units on this UnitArray.
        """
        ary = numpy.array((1,2,3))
        unit_ary = UnitArray(ary, units=dimensionless)
        self.assertEqual(unit_ary.units, dimensionless)

    def test_as_units(self):
        """ Conversion from one unit system to another.
        """
        ary = numpy.array((1,2,3))
        unit_ary = UnitArray(ary, units=meters)
        new_unit_ary = unit_ary.as_units(feet)

        # Test the values are correct
        desired_array = units.convert(ary, meters, feet)
        self.assert_(numpy.all(desired_array==new_unit_ary))

        # Also, make sure the units are correctly assigned.
        self.assertEqual(new_unit_ary.units, feet)

        # fixme: We should also test that any other items in the unit_ary are
        #        copied.


    ############################################################################
    # Test cloning type behavior.
    ############################################################################

    def test_slice_keeps_attributes(self):
        """ Does a sliced version of a UnitArray keeps its attributes?
        """
        ary = numpy.array((1,2,3))
        unit_ary = UnitArray(ary, units=meters)
        new_unit_ary = unit_ary[:2]

        # Also, make sure the units and family_name are correctly assigned.
        self.assertEqual(new_unit_ary.units, meters)

    def test_add_keeps_attributes(self):
        """ In "new=UnitArray+ary", new should have attributes of UnitArray.
        """
        ary = numpy.array((1,2,3))
        unit_ary = UnitArray(ary, units=meters)
        new_unit_ary = ary + unit_ary

        # Also, make sure the units and family_name are correctly assigned.
        self.assertEqual(new_unit_ary.units, meters)

    def test_concatenate(self):
        unit_ary_1 = UnitArray(numpy.array((1,2,3)), units=meters)
        unit_ary_2 = UnitArray(numpy.array((4,5,6)), units=meters)

        new_unit_ary = UnitArray.concatenate([unit_ary_1, unit_ary_2])
        expected = UnitArray((1,2,3,4,5,6), units=meters)
        self.assertTrue(numpy.all(new_unit_ary == expected))

    def test_concatenate_keeps_attribute(self):
        """ concatenating does not call __new__ on the result
            so the attribute must be set correctly in __array_finalize__
        """
        unit_ary_1 = UnitArray(numpy.array((1,2,3)), units=meters)
        unit_ary_2 = UnitArray(numpy.array((1,2,3)), units=meters)

        new_unit_ary = UnitArray.concatenate([unit_ary_1, unit_ary_2])

        self.assertEqual(new_unit_ary.units, meters)

    def test_concatenate_keeps_attribute_using_mixed_units(self):
        """ concatenating does not call __new__ on the result
            so the attribute must be set correctly in __array_finalize__.
            This test creates another unit array with different units
            before the concatenate to make sure the finalize method does not
            assign the wrong units
        """
        unit_ary_1 = UnitArray(numpy.array((1,2,3)), units=meters)
        unit_ary_2 = UnitArray(numpy.array((1,2,3)), units=meters)

        unit_ary_3 = UnitArray(numpy.array((1,2,3)), units=feet)
        new_unit_ary = UnitArray.concatenate([unit_ary_1, unit_ary_2])

        self.assertEqual(new_unit_ary.units, meters)

    ############################################################################
    # Test mathematical operations
    ############################################################################

    def test_index_name(self):
        """ Can set and retreive the index_name?
        """
        ary = numpy.array((1,2,3))
        unit_ary = UnitArray(ary)
        unit_ary.index_name = "depth"
        self.assertEqual(unit_ary.index_name, "depth")


    ############################################################################
    # Test Construction Speed
    ############################################################################

    @skip
    def test_construction_is_not_slow(self):
        """ Ensure instantiating a UnitArray is <1.1 slower than for arrays.

            Note that we are using a very small array to test this case so that
            we are seeing the "worst case" overhead.
        """

        ### Parameters #########################################################

        # Slowdown we will allow compared to standard python evaluation
        allowed_slowdown = 1.1

        # Number of timer iterations.
        N = 30000

        ### Array creation #####################################################
        setup = "from numpy import array\n"
        expr = 'array((1,2,3))'
        std = timeit.Timer(expr, setup)
        std_res = std.timeit(N)

        ### UnitArray creation #################################################
        setup = "from scimath.units.api import UnitArray\n"
        expr = 'UnitArray((1,2,3))'
        unit_ary = timeit.Timer(expr, setup)
        unit_ary_res = unit_ary.timeit(N)

        slowdown = unit_ary_res/std_res
        assert slowdown < allowed_slowdown, 'actual slowdown: %f' % slowdown

    @skip
    def test_finalize_is_not_slow(self):
        """ Finalizing (cloning) a UnitArray is less than 1.1 slower than arrays

            __array_finalize__ is called every time a math operation, slice,
            etc. are called on a unit_array.  We do not want this to have much
            overhead compared to stanard array operations.

            Here, we compare the speed of adding 1 to an array and to a
            UnitArray.

            Note that we are using a very small array to test this case so that
            we are seeing the "worst case" overhead.
        """

        ### Parameters #########################################################

        # Slowdown we will allow compared to standard python evaluation
        allowed_slowdown = 1.1

        # Number of timer iterations.
        N = 30000

        ### Array ##############################################################
        setup = "from numpy import array\n" \
                "val = array((1,2,3))\n"
        expr = 'val+1'
        std = timeit.Timer(expr, setup)
        std_res = std.timeit(N)

        ### UnitArray ##########################################################
        setup = (
            "from scimath.units.api import UnitArray\n"
            "val = UnitArray((1,2,3))\n"
        )
        expr = 'val+1'
        unit_ary = timeit.Timer(expr, setup)
        unit_ary_res = unit_ary.timeit(N)

        slowdown = unit_ary_res/std_res
        assert slowdown < allowed_slowdown, 'actual slowdown: %f' % slowdown

class UnitArrayPickleTestCase(unittest.TestCase):

    def test_pickle(self):
        'Pickling'
        for a in (

            UnitArray(0),
            UnitArray([0,1]),
            UnitArray([0,1], units=feet/seconds),
            UnitArray([[0.5,1.0], [10.,20.]], units=meters),

            ):
            state = dumps(a)
            b = loads(state)
            self.assertTrue(all(b == a))
            self.assertTrue(hasattr(b, 'units'))
            self.assertEqual(b.units,a.units)

class PassUnitsTestCase(unittest.TestCase):
    """ Some ufuncs keep units.
    """

    ############################################################################
    # ConvertUnitsTestCase interface.
    ############################################################################

    def test_add(self):
        a = UnitArray([1,2,3],units=meters/second)
        b = UnitArray([1,2,3],units=meters/second)
        result = a + b
        self.assertEqual(result.units, meters/second)

    def test_add_compatible(self):
        a = UnitArray([1,2,3],units=meters)
        b = UnitArray([1,2,3],units=feet)
        result = a + b
        mplusf = (meters+feet).value
        self.assertEqual(result[0], mplusf)
        self.assertEqual(result.units, meters)

    def test_add_nopass(self):
        a = UnitArray([1,2,3],units=meters/second)
        b = UnitArray([1,2,3],units=feet)
        self.assertRaises(InvalidConversion, operator.add, a, b)

    def test_subtract(self):
        a = UnitArray([1,2,3],units=meters/second)
        b = UnitArray([1,2,3],units=meters/second)
        result = a - b
        self.assertEqual(result.units, meters/second)

    def test_subtract_compatible(self):
        a = UnitArray([1,2,3],units=meters)
        b = UnitArray([1,2,3],units=feet)
        result = a - b
        mminusf = (meters-feet).value
        self.assertEqual(result[0], mminusf)
        self.assertEqual(result.units, meters)

    def test_subtract_dimensionless(self):
        a = UnitArray([1,2,3], units=dimensionless)
        b = 1
        result = a - b
        self.assertEqual(result.units, dimensionless)
        self.assertEqual(result, UnitArray([0,1,2], units=dimensionless))
        result = b - a
        self.assertEqual(result.units, dimensionless)
        self.assertEqual(result, UnitArray([0,-1,-2], units=dimensionless))
        c = array([3,2,1])
        result = a - c
        self.assertEqual(result.units, dimensionless)
        self.assertEqual(result, UnitArray([-2,0,2], units=dimensionless))
        result = c - a
        self.assertEqual(result.units, dimensionless)
        self.assertEqual(result, UnitArray([2,0,-2], units=dimensionless))

    def test_divide_pass(self):
        a = UnitArray([1,2,3],units=meters/second)
        result = a/3.0
        self.assertEqual(result.units, meters/second)
        result = 3.0/a
        self.assertEqual(result.units, second/meters)
        b = UnitArray([3,2,1], units=second)
        result = a/b
        self.assertEqual(result.units, meters/second**2)

    def test_multiply_pass(self):
        a = UnitArray([1,2,3],units=meters/second)
        result = a*3.0
        self.assertEqual(result.units, meters/second)
        result = 3.0*a
        self.assertTrue((array(result) == array([3.0,6.0,9.0])).all())
        self.assertEqual(result.units, meters/second)
        result = 10*a
        self.assertEqual(result.units, meters/second)
        result = a*10
        self.assertTrue((array(result) == array([10,20,30])).all())
        self.assertEqual(result.units, meters/second)
        result = UnitScalar(3.0, units=second)*a
        self.assertTrue((array(result) == array([3.0,6.0,9.0])).all())
        self.assertEqual(result.units, meters)


    def test_multiply_units(self):
        a = UnitArray([1,2,3],units=meters)
        b = UnitArray([1,2,3],units=second)
        result = a*b
        self.assertEqual(result.units, meters*second)

    def test_pow_pass(self):
        a = UnitArray([1,2,3],units=meters)
        b = 0.5
        result = a**b
        self.assertEqual(result.units, meters**0.5)
        c = array(0.5)
        result = a**c
        self.assertEqual(result.units, meters**0.5)
        d = UnitArray(0.5, units=dimensionless)
        result = a**d
        self.assertEqual(result.units, meters**0.5)

    def test_sqrt_no_pass(self):
        a = UnitArray([1.0,2.0,3.0], units=meters/second)
        result = sqrt(a)
        self.assert_(result.units is None)

    def test_sqrt_pass(self):
        a = UnitArray([1.0,2.0,3], units=dimensionless)
        result = sqrt(a)
        self.assertEqual(result.units, dimensionless)

    def test_le(self):
        a = UnitArray([1.0,2.0,3], units=dimensionless)
        b = UnitArray([3.0,1.0,1], units=2.0*dimensionless)
        result = a <= b

        # FIXME: these tests fail
        raise SkipTest

        self.assertEqual(result[0], True)
        self.assertEqual(result[1], True)
        self.assertEqual(result[2], False)
        c = 2.0
        result = b <= c
        self.assertEqual(result[0], False)
        self.assertEqual(result[1], True)
        self.assertEqual(result[2], True)
        self.assertEqual(result[0], False)
        self.assertEqual(result[1], True)
        self.assertEqual(result[2], True)

    def test_lt(self):
        a = UnitArray([1.0,4.0,3], units=dimensionless)
        b = UnitArray([3.0,2.0,1], units=2.0*dimensionless)
        result = a < b

        # FIXME: these tests fail
        raise SkipTest

        self.assertEqual(result[0], True)
        self.assertEqual(result[1], False)
        self.assertEqual(result[2], False)
        c = 4.0
        result = b < c
        self.assertEqual(result[0], False)
        self.assertEqual(result[1], False)
        self.assertEqual(result[2], True)
        d = array([1.0, 4.0, 3.0])
        result = b < d
        self.assertEqual(result[0], False)
        self.assertEqual(result[1], False)
        self.assertEqual(result[2], True)

    def test_ge(self):
        a = UnitArray([1.0,4.0,3], units=dimensionless)
        b = UnitArray([3.0,2.0,1], units=2.0*dimensionless)
        result = a >= b

        # FIXME: these tests fail
        raise SkipTest

        self.assertEqual(result[0], False)
        self.assertEqual(result[1], True)
        self.assertEqual(result[2], True)
        c = 4.0
        result = b >= c
        self.assertEqual(result[0], True)
        self.assertEqual(result[1], True)
        self.assertEqual(result[2], False)
        d = array([1.0, 4.0, 3.0])
        result = b >= d
        self.assertEqual(result[0], True)
        self.assertEqual(result[1], True)
        self.assertEqual(result[2], False)

    def test_gt(self):
        a = UnitArray([1.0,4.0,3], units=dimensionless)
        b = UnitArray([3.0,2.0,1], units=2.0*dimensionless)
        result = a > b

        # FIXME: these tests fail
        raise SkipTest

        self.assertEqual(result[0], False)
        self.assertEqual(result[1], False)
        self.assertEqual(result[2], True)
        c = 4.0
        result = b > c
        self.assertEqual(result[0], True)
        self.assertEqual(result[1], False)
        self.assertEqual(result[2], False)
        d = array([1.0, 4.0, 3.0])
        result = b > d
        self.assertEqual(result[0], True)
        self.assertEqual(result[1], False)
        self.assertEqual(result[2], False)

    def test_eq(self):
        a = UnitArray([1.0,4.0,3], units=dimensionless)
        b = UnitArray([3.0,2.0,1], units=2.0*dimensionless)
        result = a == b

        # FIXME: these tests fail
        raise SkipTest

        self.assertEqual(result[0], False)
        self.assertEqual(result[1], True)
        self.assertEqual(result[2], False)
        c = 4.0
        result = b == c
        self.assertEqual(result[0], False)
        self.assertEqual(result[1], True)
        self.assertEqual(result[2], False)
        d = array([1.0, 4.0, 3.0])
        result = b == d
        self.assertEqual(result[0], False)
        self.assertEqual(result[1], True)
        self.assertEqual(result[2], False)

    def test_ne(self):
        a = UnitArray([1.0,4.0,3], units=dimensionless)
        b = UnitArray([3.0,2.0,1], units=2.0*dimensionless)
        result = a != b

        # FIXME: these tests fail
        raise SkipTest

        self.assertEqual(result[0], True)
        self.assertEqual(result[1], False)
        self.assertEqual(result[2], True)
        c = 4.0
        result = b != c
        self.assertEqual(result[0], True)
        self.assertEqual(result[1], False)
        self.assertEqual(result[2], True)

        d = array([1.0, 4.0, 3.0])
        result = b != d
        self.assertEqual(result[0], True)
        self.assertEqual(result[1], False)
        self.assertEqual(result[2], True)

if __name__ == '__main__':
    import sys
    unittest.main(argv=sys.argv)



# -*- coding: utf-8 -*-
"""natu documentation build configuration file, created by
sphinx-quickstart on Mon Oct 15 09:06:21 2012.

This file is execfile()d with the current directory set to its containing dir.

Note that not all possible configuration values are present in this
autogenerated file.

All configuration values have a default; values that are commented out serve to
show the default.
"""

# Standard pylint settings for this project:
# pylint: disable=I0011, C0302, C0325, R0903, R0904, R0912, R0913, R0914, R0915,
# pylint: disable=I0011, W0141, W0142

# Other:
# pylint: disable=I0011, C0103, W0621, W0622


def skip(app, what, name, obj, skip, options):
    """Include otherwise hidden methods.
    """
    # pylint: disable=I0011, W0613
    if name in ["__call__", "__contains__", "__getitem__", "__len__"]:
        return False
    return skip


def setup(app):
    """Add roles and javascripts.
    """
    app.connect("autodoc-skip-member", skip)
    app.add_javascript('copybutton.js')
    app.add_javascript('analytics.js')

# -- General configuration -----------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.doctest',
              'sphinx.ext.pngmath',
              'sphinx.ext.intersphinx',
              'sphinx.ext.inheritance_diagram',
              'sphinx.ext.graphviz',
              'matplotlib.sphinxext.plot_directive'
             ]
# Note: sphinx.ext.autosummary produces a table, but I'd rather have a bulleted
# list.

# Mappings to other packages
# TODO: Fix numpy link.
intersphinx_mapping = {
    'python': ('https://docs.python.org/', None),
    'numpy': ('http://docs.scipy.org/doc/numpy/', None),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'natu'
copyright = ('2013-2014, Kevin Davies, Hawaii Natural Energy Institute and '
             'Georgia Tech Research Corporation')

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
import natu
version = natu.__version__
if not version:
    version = ''
# The full version, including alpha/beta/rc tags.
release = version

# List of documents that shouldn't be included in the build.
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# -- Options for HTML output ---------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for a
# list of builtin themes.
html_theme = 'default'

html_theme_options = {
    'stickysidebar': True,
    'sidebarbgcolor': '#888888',
    'sidebartextcolor': 'white',
    'sidebarlinkcolor': 'white',
}

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = "Natural units in Python"

# A shorter title for the navigation bar.  Default is the same as html_title.
version_str = " v%s" % version if version else ''
html_short_title = project + version_str + " documentation"

# The name of an image file (relative to this directory) to place at the top of
# the sidebar.
html_logo = '_static/logo.svg'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = '_static/favicon.ico'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files, so
# a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']
html_style = 'custom.css'

# List of paths that contain extra files not directly related to the
# documentation
html_extra_path = ['extra']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
html_sidebars = {'**': ['info.html', 'definitions.html', 'user.html',
                        'api.html', 'links.html', 'searchbox.html',
                        'download.html']}

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
html_use_opensearch = 'False'

# Output file base name for HTML help builder.
htmlhelp_basename = 'natuDoc'

math_output = 'MathML'




from unittest import TestCase

from scimath.physical_quantities.dimensions import Dimensions, Dim

class DimensionsTest(TestCase):
    def setUp(self):
        self.dimensionless = Dimensions({})
        self.length = Dimensions({"length": 1.0})
        self.mass = Dimensions({"mass": 1.0, 'time': 0.0})
        self.time = Dimensions({"time": 1.0})
        self.acceleration = Dimensions({"length": 1.0, "time": -2.0})
        self.force = Dimensions({"mass": 1.0, "length": 1.0, "time": -2.0})
        self.also_force = Dimensions({"mass": 1.0, "length": 1.0,
                                              "time": -2.0})

    # Tests for basic initialization
    def test_setup_simple(self):
        assert self.length.dimension_dict == {"length": 1.0}

    def test_setup_remove_dimensions(self):
        assert self.mass.dimension_dict == {"mass": 1.0}

    # Tests for expansion property
    def test_expansion_dimensionless(self):
        assert self.dimensionless.expansion == "dimensionless"

    def test_expansion_simple(self):
        assert self.length.expansion == "length"

    def test_expansion_complex(self):
        assert self.force.expansion == "length*mass*time**-2"

    def test_expansion_expression(self):
        velocity = self.length/self.time
        assert velocity.expansion == "length*time**-1"

    def test_expansion_complete_cancellation(self):
        dimensionless = self.force/self.force
        assert dimensionless.expansion == "dimensionless"

    # Tests for __str__ method
    def test_str_dimensionless(self):
        assert str(self.dimensionless) == "dimensionless"

    def test_str_simple(self):
        assert str(self.length) == "length"

    def test_str_complex(self):
        assert str(self.force) == "length*mass*time**-2"

    def test_str_expression(self):
        velocity = self.length/self.time
        assert str(velocity) == "length*time**-1"

    # Tests for equality and inequality
    def test_equality(self):
        assert self.force == self.also_force

    def test_inequality(self):
        assert self.force != self.acceleration

    # Tests for hashing
    def test_hash(self):
        assert hash(self.force) == hash(self.also_force)

    def test_not_hash(self):
        # there is a very small chance that this will fail when it should not
        # because of a hash collision
        assert hash(self.force) != hash(self.acceleration)

    # Tests for arithmetic operations
    def test_mul(self):
        force = self.acceleration*self.mass
        assert force.dimension_dict == {"length": 1.0, "time": -2.0, "mass": 1.0}

    def test_div(self):
        velocity = self.length/self.time
        assert velocity.dimension_dict == {"length": 1.0, "time": -1.0}

    def test_pow(self):
        area = self.length**2
        assert area.dimension_dict == {"length": 2.0}

    def test_cancellation(self):
        acceleration = self.force*self.time/self.mass
        assert acceleration.dimension_dict == {"length": 1.0, "time": -1.0}

    def test_complete_cancellation(self):
        dimensionless = self.force/self.force
        assert dimensionless.dimension_dict == {}



# Standard Library imports
import unittest

# Enthought Library imports
from traits.testing.api import doctest_for_module

geo_imports = [
    'GPA', 'GPa', 'Gpa', 'MPA', 'MPa', 'MPa_per_100f', 'MPa_per_100ft',
    'MPa_per_f', 'MPa_per_ft', 'MPa_per_m', 'Mpa', 'N', 'ampere', 'api',
    'apsi', 'atto', 'bar', 'bars', 'becquerel', 'candela', 'centi', 'copy',
    'coulomb', 'cubic_centimeter', 'cubic_foot', 'cubic_meter', 'deci', 'deka',
    'dimensionless', 'exa', 'farad', 'femto', 'foot', 'frac', 'fraction',
    'fractional', 'g_ft_per_cc_s', 'g_km_per_cc_s', 'gapi', 'giga', 'gpa',
    'grams', 'gray', 'hecto', 'henry', 'hertz', 'inch', 'joule', 'katal',
    'kbar', 'kbars', 'kilo', 'kilobar', 'kilogram', 'kilometers', 'lb',
    'lb_per_gal', 'lb_per_gallon', 'lbf', 'lbs', 'liter', 'lumen', 'lux', 'm',
    'mS', 'mSiemen', 'mega', 'meter', 'mho', 'micro', 'microsecond', 'milli',
    'millivolts', 'mmho', 'mole', 'mpa', 'mv', 'nano', 'newton', 'none', 'ohm',
    'ohm_m', 'ohm_meter', 'ohmm', 'ohms', 'ohms_per_m', 'ohms_per_meter',
    'parts_per_million', 'parts_per_one', 'pascal', 'pct', 'percent',
    'percentage', 'peta', 'pico', 'pounds_per_square_inch', 'ppg', 'ppm',
    'psi', 'psi_per_f', 'psi_per_ft', 'psig', 'radian', 'ratio', 'second',
    'siemen', 'siemens', 'siemens_per_m', 'siemens_per_meter', 'sievert',
    'steradian', 'tera', 'tesla', 'unit', 'us_fluid_gallon', 'us_per_ft', 'v',
    'volt', 'volts', 'watt', 'weber', 'yocto', 'yotta', 'zepto', 'zetta'
    ]

def write_geo_err_msg(name):
    """write the message describing the import not found in geo_units"""
    msg = """Could not find previously available import '{0}' in 
    scimath.units.geo_units module. Check to ensure that '{0}' is
    available as an import from scimath.units.geounits."""
    return msg.format(name)

class GeoUnitsImportsTestCase(unittest.TestCase):
    """ When geo_units was initially set up, many units were defined there
    which should have been defined elsewhere and imported to geo_units, causing
    some problems later when things like "psi" could not be imported from
    scimath.units.pressure but were instead found in
    scimath.units.geo_units. In Dec 2011, the units were reorganized to make
    units available in their expected location. This test was set up to ensure
    that not code that imported a unit from geo_units would break in the
    reorganization.
    
    """
    def test_geo_imports(self):
        """ Test whether everything that was availabile in geo_units still is available.
        """
        from scimath.units import geo_units

        available_now = geo_units.__dict__
        
        for name in geo_imports:
            self.assertIn(name, available_now, msg=write_geo_err_msg(name))
            
if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
"""Decorators to update the functions of Python math and Numpy umath
"""

# pylint: disable=I0011, C0103, C0111, E0611, E1101, W0141, W0142, W0621

from sys import version
from . import core
from .core import (Quantity, ScalarUnit, assert_homogeneous, homogeneous,
                    value, dimensionless_value)
from .units import rad

# TODO: Combine some of these once Python can propagate a function's signature
# up through a decorator (PEP 362?).

# functools
# ---------

if version >= '3':
    from functools import wraps
else:
    # Python2's functools.wraps requires that all of the attributes are
    # present.  We need "except AttributeError: pass" to wrap numpy ufuncs,
    # since they don't have __module__.

    from functools import WRAPPER_ASSIGNMENTS, WRAPPER_UPDATES, partial

    # The following two functions are from Python 3.2
    # (http://hg.python.org/cpython/file/3.2/Lib/functools.py).

    def update_wrapper(wrapper, wrapped,
                       assigned=WRAPPER_ASSIGNMENTS,
                       updated=WRAPPER_UPDATES):
        """Update a wrapper function to look like the wrapped function.

        **Parameters:**

        - :func:`wrapper`: function to be updated

        - :func:`wrapped`: original function

        - *assigned*: tuple naming the attributes assigned directly from
          :func:`wrapped` to :func:`wrapper`

             *assigned* defaults to :const:`functools.WRAPPER_ASSIGNMENTS`.

        - *updated*: tuple naming the attributes of :func:`wrapper` that are
          updated with the corresponding attribute from :func:`wrapped`

             *updated* defaults to :const:`functools.WRAPPER_UPDATES`.
        """
        wrapper.__wrapped__ = wrapped
        for attr in assigned:
            try:
                value = getattr(wrapped, attr)
            except AttributeError:
                pass
            else:
                setattr(wrapper, attr, value)
        for attr in updated:
            getattr(wrapper, attr).update(getattr(wrapped, attr, {}))

        # Return the wrapper so this can be used as a decorator via partial().
        return wrapper

    def wraps(wrapped,
              assigned=WRAPPER_ASSIGNMENTS,
              updated=WRAPPER_UPDATES):
        """Apply :func:`update_wrapper` to a wrapper function.

        This decorator factory returns a decorator that invokes
        :func:`update_wrapper` with the decorated function as the wrapper
        argument and the arguments to this function as the remaining arguments.
        Default arguments are as for :func:`update_wrapper`.  This is a
        convenience function to simplify applying :func:`partial` to
        :func:`update_wrapper`.
        """
        return partial(update_wrapper, wrapped=wrapped,
                       assigned=assigned, updated=updated)

# Standard functions
# ------------------


def merge(value, prototype):
    """Merge *value* into a new :class:`~natu.core.ScalarUnit` or
    :class:`~natu.core.Quantity` with the properties (:attr:`dimension`,
    :attr:`display`, etc.) of *prototype*.

    If *prototype* is not a :class:`~natu.core.ScalarUnit` or
    :class:`~natu.core.Quantity`, then return *value* directly.
    """
    if isinstance(prototype, ScalarUnit):
        return ScalarUnit(value, prototype.dimension, prototype.display,
                          prototype.prefixable)
    return core.merge(value, prototype)


def merge_raise(value, prototype, power):
    """Same as :func:`merge`, but raise the dimension and display unit to
    *power*
    """
    if isinstance(prototype, ScalarUnit):
        return ScalarUnit(value, power * prototype.dimension,
                          power * prototype.display, prototype.prefixable)
    try:
        return Quantity(value, power * prototype.dimension,
                        power * prototype.display)
    except AttributeError:
        return value

# Elementary wrappers
# -------------------


def arg_x(func):
    """Decorate a function to accept *x* as a keyword argument.
    """
    @wraps(func)
    def wrapped(x):
        return func(x)
    return wrapped


def arg_xi(func):
    """Decorate a function to accept *x* and *i* as keyword arguments.
    """
    @wraps(func)
    def wrapped(x, i):
        return func(x, i)
    return wrapped


def arg_yx(func):
    """Decorate a function to accept *y* and *x* as keyword arguments.
    """
    @wraps(func)
    def wrapped(y, x):
        return func(y, x)
    return wrapped


def change_doc(func, doc=None):
    """Update the docstring of a function.

    Wrap the function if the docstring is not writeable.
    """
    if doc is None:
        return func
    try:
        func.__doc__ = doc
    except AttributeError:
        @wraps(func)
        def wrapped(x):
            """Pass-through function of x"""
            return func(x)
        wrapped.__doc__ = doc
        return wrapped
    return func


def copy_props(func):
    """Decorate a function to return a :class:`~natu.core.ScalarUnit` or
    :class:`~natu.core.Quantity` instance that matches the first argument except
    for the computed value.
    """
    @wraps(func)
    def wrapped(x, *args, **kwargs):
        return merge(func(x, *args, **kwargs), x)

    wrapped.__doc__ += (
        "\nThe properties of the first term (:attr:`dimension`, \n"
        "\n:attr:`display`, etc.) are copied to the result.\n")
    return wrapped


def dimensionless(func):
    """Decorate a function to accept a number or a dimensionless quantity.

    If the quantity is not dimensionless, a TypeError is raised.
    """
    @wraps(func)
    def wrapped(*args):
        return func(*map(dimensionless_value, args))

    return wrapped


def dimensionless_implicit(func):
    """Decorate a function to give a special :class:`TypeError` message about
    dimensionality.

    This is triggered when a quantity is cast as a :class:`float` or
    :class:`int` but it is not dimensionless.
    """
    @wraps(func)
    def wrapped(*args):
        try:
            return func(*args)
        except TypeError:
            raise TypeError("The quantity must be dimensionless.")

    return wrapped

# Compound wrappers
# -----------------


def dimensionless_copy_props(func, doc=None):
    """Wrap a function to accept dimensionless quantities and pass the
    dimension, display unit, etc.  Change the docstring to *doc*.
    """
    return change_doc(arg_x(copy_props(dimensionless_implicit(func))), doc)


def trig(func, doc=None):
    """Wrap a trigonometric function that accepts angle in radians to accept
    angle as a quantity.
    """
    @wraps(func)
    def wrapped(theta):
        try:
            return func(theta / rad)
        except TypeError:
            raise TypeError("The argument must be an angle or zero.")

    return change_doc(wrapped, doc)


def inv_trig(func, doc=None):
    """Wrap an inverse trigonometric function that returns angle in radians to
    return angle as a quantity.
    """
    @wraps(func)
    def wrapped(x):
        return func(x) * rad

    return change_doc(wrapped, doc)


def inv_trig_yx(func, doc=None):
    """Wrap a dual-input inverse trigonometric function that returns angle in
    radians to return angle as a quantity.
    """
    # Note that *args isn't used because the variables wouldn't be labeled in
    # the documentation.
    @arg_yx
    @homogeneous
    @wraps(func)
    def wrapped(y, x):
        return func(y, x) * rad

    return change_doc(wrapped, doc)


def homogeneous_copy_props(func):
    """Wrap a function to use the :attr:`value` of quantities and pass the
    dimension, display unit, etc. of the first argument.  Check that the
    quantities have the same dimension.
    """
    return copy_props(homogeneous(func))


def homogeneous_copy_props_iter(func):
    """Decorate a function to use the values of an iterable of quantities and
    pass the dimension, display unit, etc. of the first quantity.
    """
    @wraps(func)
    def wrapped(x):
        assert_homogeneous(*x)
        return merge(func(map(value, x)), x[0])

    return wrapped


def use_value(func, doc=None):
    """Wrap a function to use the value of a quantity.
    """
    @wraps(func)
    def wrapped(x):
        return func(value(x))

    return change_doc(wrapped, doc)


def use_value_copy_props(func, doc):
    """Wrap a function to use the :attr:`value` of a quantity and pass the
    dimension, display unit, etc.  Change the docstring to *doc*.
    """
    return change_doc(arg_x(copy_props(use_value(func))), doc)


def use_value_raise(func, y=None):
    """Wrap a function to use the value of a quantity and raise the dimension
    and display unit to a power.

    If the power (*y*) is not given as an argument to this wrapper, then it is
    an argument to the function itself.
    """
    @wraps(func)
    def wrapped(x, y):
        return merge_raise(func(value(x), y), x, y)

    @wraps(func)
    def wrapped_fixed(x):
        return merge_raise(func(value(x)), x, y)

    return wrapped if y is None else wrapped_fixed


def use_values_copy_props(func, doc=None):
    """Decorate a function to use the values of two quantities and pass the
    properties (:attr:`dimension`, :attr:`display`, etc.) of the first.
    """
    @copy_props
    @wraps(func)
    def wrapped(x, y):
        return func(value(x), value(y))

    return change_doc(wrapped, doc)


def use_values_copy_props_xi(func, doc=None):
    """Wrap a function to use the :attr:`value` of quantities and pass the
    dimension, display unit, etc. of the first argument.  Change the docstring
    to *doc*.
    """
    return change_doc(arg_xi(use_values_copy_props(func)), doc)



from numpy.testing import assert_equal
from scimath.units.unit_scalar import UnitScalar
from scimath.units.unit import dimensionless

def dimensionless_test():
    """
    Test the modification to the division, multiplication and pow
    such that a dimensionless quantity formed by is indeed dimensionless
    """


    a = UnitScalar(1.0, units = 'm')
    b = UnitScalar(2.0, units = 'mm')
    d = UnitScalar(2.0, units = 'm**(-1)')

    c = a/b
    e = b*d

    f= UnitScalar(2.0, units = dimensionless)
    g = f**2

    assert_equal(c.units,dimensionless)
    assert_equal(e.units,dimensionless)
    assert_equal(g.units,dimensionless)

